import inspect
import logging
import statistics
import threading
import time
from contextlib import contextmanager
from typing import Any, Generator, Optional, Union

from jf_ingest.logging_helper import send_to_agent_log_file
from jf_ingest.telemetry.metrics import JellyCounter, JellyGauge, JellyHistogram


class AdaptiveThrottler:
    """
    Thread safe adaptive throttler that dynamically applies and adjusts backoff times based
    on initial request durations, percentile threshold and backoff/reverse backoff factors.
    """

    def __init__(
        self,
        max_rps: float = 10.0,
        baseline_window_size: int = 50,
        percentile_threshold: float = 0.95,
        backoff_factor: float = 0.90,
        reverse_backoff_factor: float = 0.95,
        reverse_backoff_counter: int = 25,
        max_backoff_time: float = 60.0,
        logging_extra: Optional[dict[str, str]] = None,
    ) -> None:
        """
        Initialize the adaptive throttler with the specified parameters.

        Args:
            max_rps: Maximum request rate allowed (requests per second).
            baseline_window_size: Number of response times to collect before establishing the baseline.
            percentile_threshold: The percentile of the baseline response times to use as the threshold.
            backoff_factor: Factor by which to decrease the request rate when response times exceed the threshold.
            reverse_backoff_factor: Factor by which to increase the request rate when response times are
                                    consistently below the threshold
            reverse_backoff_counter: Number of consecutive responses within the threshold to trigger a slight
                                     relaxation of the backoff factor.
            max_backoff_time: Maximum backoff time to apply when the response time exceeds the threshold (in seconds).
            logging_extra: Extra logging information to include in the log messages
        """

        # Max RPS is the maximum request rate allowed. If the baseline RPS is higher than this,
        # throttle based on the max_rps instead. This ensures that the minimum backoff time is
        # significant enough to prevent overloading the server.
        self.max_rps = max_rps

        # Backoff factor is the degree to which the backoff rate is increased when the response
        # time exceeds the threshold. For example, if the backoff factor is 0.75, the backoff rate
        # will be increased by 25% each time the response time exceeds the threshold.
        self.backoff_factor = backoff_factor

        # Reverse backoff factor is the degree to which the backoff rate is decreased when the
        # response time is consistently below the threshold. For example, if the reverse backoff
        # factor is 0.95, the backoff rate will be decreased by 5% each time the response time is
        # consistently below the threshold
        self.reverse_backoff_factor = reverse_backoff_factor

        # Reverse backoff counter is the number of consecutive responses within the threshold that
        # will trigger a slight relaxation of the backoff factor. This is useful for gradually
        # increasing the request rate when (inferred) server load returns to normal.
        self.reverse_backoff_counter = reverse_backoff_counter

        # Max backoff time is the maximum backoff time to apply when the response time exceeds the threshold.
        # This is used to prevent the backoff time from becoming too large and causing the throttler to
        # wait for an excessive amount of time.
        self.max_backoff_time = max_backoff_time

        # Window size is the number of response times to collect before establishing the baseline.
        # A larger window size will provide a more accurate baseline, but will take longer to
        # establish and is at risk of treating quick increases in server load as normal.
        # A smaller window size will establish the baseline more quickly, but may be less accurate.
        self._window_size = baseline_window_size

        # Percentile threshold is the percentile of the baseline response times to use as the
        # incoming response time threshold.
        self._percentile_threshold = percentile_threshold

        # Request rate is the dynamically adjusted request rate based on the baseline response times.
        # Backoff times are calculated based on this rate. This variable is adjusted based on the backoff
        # factor and reverse backoff factor when the response time exceeds or is below the threshold.
        self._request_rate: float = 0.0

        # Initial request rate is the initial request rate calculated based on the baseline response.
        # This variable does not change once set, and is used for comparison with the dynamically adjusted
        # request rate.
        self._initial_request_rate: float = 0.0

        # Percentile threshold time is the calculated response time at the specified percentile of the
        # baseline response times. This is used as the threshold for incoming response times.
        self._percentile_threshold_time: float = 0.0

        # Response times is a list of response times used to establish the baseline.
        self._response_times: list[float] = []

        # Response within threshold counter is the counter of consecutive recorded responses within
        # the threshold.
        self._response_within_threshold_counter = 0

        self.logging_extra: dict[str, Any] = {}
        self.metrics_attrs: dict[str, Any] = {}

        if logging_extra:
            self.logging_extra = logging_extra

            for k in ["company_slug", "function_name"]:
                if v := logging_extra.get(k):
                    self.metrics_attrs = {k: v}

        self._lock = threading.Lock()

        # Initialize metrics
        self._response_time_histogram = JellyHistogram(
            name="jf_ingest_adaptive_throttler_response_times",
            unit="s",
            description="Response times observed by AdaptiveThrottler",
        )

        self._request_rate_gauge = JellyGauge(name="jf_ingest_adaptive_throttler_request_rate")

        self._backoff_counter = JellyCounter(
            name="jf_ingest_adaptive_throttler_backoff_events",
            description="Number of times backoff was applied in AdaptiveThrottler",
        )

        self._reverse_backoff_counter = JellyCounter(
            name="jf_ingest_adaptive_throttler_reverse_backoff_events",
            description="Number of times reverse backoff was applied in AdaptiveThrottler",
        )

        self._backoff_time_histogram = JellyHistogram(
            name="jf_ingest_adaptive_throttler_backoff_times",
            unit="s",
            description="Backoff times applied by AdaptiveThrottler",
        )

        self._total_requests_counter = JellyCounter(
            name="jf_ingest_adaptive_throttler_total_requests",
            description="Total number of requests processed by AdaptiveThrottler",
        )

        self._log_attrs()

    def _log_attrs(self) -> None:
        """Logs the relevant object attributes."""
        attrs: dict[str, Union[float, int]] = {}

        for attr, val in self.__dict__.items():
            if isinstance(val, (float, int)):
                attrs[attr] = val

        send_to_agent_log_file(
            f"Initialized adaptive throttler with parameters: {attrs}", extra=self.logging_extra
        )

    def _record_response_time(self, response_time: float) -> None:
        """Record response time for establishing the baseline."""
        self._response_times.append(response_time)

        if len(self._response_times) == self._window_size:
            self._compute_percentile()
            self._calculate_dynamic_rate()

    def _compute_percentile(self) -> None:
        """Compute the specified percentile of the baseline response times."""
        sorted_times = sorted(self._response_times)
        i = int(self._percentile_threshold * len(sorted_times))
        percentile_value = sorted_times[i]

        # Calculate standard deviation
        std_dev = statistics.stdev(self._response_times)

        # Adjust threshold to be 2 standard deviations above the percentile
        self._percentile_threshold_time = percentile_value + (2 * std_dev)

        msg_prefix = (
            f"{self._percentile_threshold * 100:.0f}th percentile response time with adjustment"
        )
        send_to_agent_log_file(
            f"{msg_prefix}: {percentile_value:.4f} + 2σ ({2 * std_dev:.4f}) = {self._percentile_threshold_time:.4f} seconds",
            extra=self.logging_extra,
        )

    def _calculate_dynamic_rate(self) -> None:
        """Calculate the initial request rate dynamically based on the baseline response times."""
        average_time = sum(self._response_times) / len(self._response_times)
        rr = min(1 / average_time, self.max_rps)
        self._request_rate, self._initial_request_rate = rr, rr
        send_to_agent_log_file(
            f"Dynamic initial request rate calculated: {rr:.4f} rps",
            extra=self.logging_extra,
        )
        self._request_rate_gauge.measure(value=self._request_rate, attributes=self.metrics_attrs)

    @contextmanager
    def process_response_time(self) -> Generator[None, None, None]:
        """
        Context manager that records the duration of code execution within the 'with' block,
        processes the response time, and applies backoff if required.

        Usage:
            with adaptive_throttler.process_response_time():
                # Code to execute and measure
        """
        if not self.metrics_attrs.get("function_name"):
            current_frame = inspect.currentframe()

            if current_frame and current_frame.f_back and current_frame.f_back.f_back:
                caller_frame = current_frame.f_back.f_back
                f_name = caller_frame.f_code.co_name
                send_to_agent_log_file(
                    f"Adaptive throttler called from {f_name}", extra=self.logging_extra
                )
            else:
                send_to_agent_log_file(
                    "Adaptive throttler called from unknown function",
                    level=logging.WARNING,
                    extra=self.logging_extra,
                )
                f_name = "unknown"

            self.metrics_attrs["function_name"] = f_name
            self.logging_extra["function_name"] = f_name

        start_time = time.perf_counter()

        try:
            yield
        finally:
            end_time = time.perf_counter()
            response_time = end_time - start_time

            with self._lock:
                self._response_time_histogram.measure(
                    value=response_time,
                    attributes=self.metrics_attrs,
                )
                self._total_requests_counter.add(amount=1, attributes=self.metrics_attrs)

                if len(self._response_times) < self._window_size:
                    self._record_response_time(response_time)
                    return

                # If the response time exceeds the threshold, decrease the request rate
                # and apply the backoff time
                if (
                    self._percentile_threshold_time
                    and response_time > self._percentile_threshold_time
                ):
                    send_to_agent_log_file(
                        f"Response time exceeded threshold: {response_time:.4f} > {self._percentile_threshold_time:.4f}",
                        extra=self.logging_extra,
                    )

                    self._request_rate *= self.backoff_factor
                    self._response_within_threshold_counter = 0
                    backoff_time = 1 / self._request_rate

                    if backoff_time > self.max_backoff_time:
                        send_to_agent_log_file(
                            f'Backoff time {backoff_time} exceeded max backoff time {self.max_backoff_time}',
                            level=logging.WARNING,
                            extra=self.logging_extra,
                        )
                        backoff_time = self.max_backoff_time

                    send_to_agent_log_file(
                        f"Request rate decreased to {self._request_rate:.4f} rps, waiting {backoff_time:.4f} seconds",
                        extra=self.logging_extra,
                    )

                    self._request_rate_gauge.measure(
                        value=self._request_rate,
                        attributes=self.metrics_attrs,
                        record_immediate=True,
                    )
                    self._backoff_counter.add(
                        amount=1, attributes=self.metrics_attrs, record_immediate=True
                    )
                    self._backoff_time_histogram.measure(
                        value=backoff_time, attributes=self.metrics_attrs, record_immediate=True
                    )

                    time.sleep(backoff_time)
                elif self._request_rate < self._initial_request_rate:
                    self._response_within_threshold_counter += 1

                    # If there have been enough consecutive responses within the threshold, slightly
                    # relax the backoff factor
                    if self._response_within_threshold_counter == self.reverse_backoff_counter:
                        send_to_agent_log_file(
                            f"Response time within threshold for {self.reverse_backoff_counter} consecutive responses",
                            extra=self.logging_extra,
                        )

                        self._request_rate = min(
                            self._request_rate / self.reverse_backoff_factor,
                            self._initial_request_rate,
                        )
                        self._response_within_threshold_counter = 0

                        self._request_rate_gauge.measure(
                            value=self._request_rate,
                            attributes=self.metrics_attrs,
                            record_immediate=True,
                        )
                        self._reverse_backoff_counter.add(
                            amount=1, attributes=self.metrics_attrs, record_immediate=True
                        )

                        send_to_agent_log_file(
                            f"Request rate increased to {self._request_rate:.4f} rps",
                            extra=self.logging_extra,
                        )



import asyncio
import functools
import inspect
import logging
import sys
from contextlib import contextmanager
from functools import partial
from typing import Generator, Optional

from opentelemetry import baggage, context, trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor, SpanExporter

logger = logging.getLogger(__name__)


def init_opentelemetry_tracer(exporter: SpanExporter) -> None:

    trace_provider = TracerProvider()
    processor = BatchSpanProcessor(exporter)
    trace_provider.add_span_processor(processor)

    # Sets the global default tracer provider
    trace.set_tracer_provider(trace_provider)


def add_telemetry_fields(fields) -> None:
    for key, val in fields.items():
        add_telemetry_span_field(key, val)
        add_telemetry_trace_field(key, val)


def add_telemetry_trace_field(key: str, value) -> None:
    """
    Adds the given key and value to OTel Baggage.

    Use this function when you want to add a context field/attribute
    to the entire trace (not just the current span.)
    """

    # NOTE(asm,2023-06-07): the opentelemetry documentation claims that you need to keep track of
    # the token returned from `context.attach` and call `detach` at the conclusion of the span. This
    # is difficult to wrangle, particularly in the case of decorated functions. Throwing caution to
    # the wind, we're going to just call `context.attach` and not `detach`. This doesn't _seem_ to
    # be causing any problems at the time of writing, but if we start getting incoherent-looking
    # tracing data, this may be why. A better solution might be to create our own decorator for
    # tracing functions, but at that point we'd basically be writing our own telemetry library and
    # I'd like to avoid that (for now, at least).
    # See: https://docs.honeycomb.io/getting-data-in/opentelemetry/python-distro/#multi-span-attributes
    context.attach(baggage.set_baggage(key, value))


def add_telemetry_span_field(key, value):
    """
    Adds the given key and value to both the OTel attribute
    for the current span.

    Use this function when you want to add a context field/attribute
    to the current span, but not the whole trace.

    """
    set_current_span_attribute(key=key, value=value)


def add_telemetry_span_fields(fields: dict):
    """
    Adds the given key/value pairs to the OTel attribute for the current span.

    Use this function when you want to add context fields/attributes
    to the current span, but not the whole trace.
    """
    for key, val in fields.items():
        add_telemetry_span_field(key=key, value=val)


def set_current_span_attribute(key: str, value):
    """
    Sets an attribute  with a key value for the current span.
    """

    span = trace.get_current_span()
    span.set_attribute(key, value)


@contextmanager
def record_span(name: str, attributes: Optional[dict] = None) -> Generator[None, None, None]:
    """
    Start a new span with the given attributes. This is a convenience function that wraps
    the start_span and finish_span/end functions from OTel.
    The span will be automatically finished/ended when the context manager exits.
    We use a convention where the calling module's name is used as the tracer name.
    The given name will be used as the span name.
    """
    try:
        # Get the current frame and check if it's not None
        current_frame = inspect.currentframe()
        if current_frame is None or current_frame.f_back is None:
            raise AttributeError("Could not get the previous frame for tracing")

        # Get the module of the calling function and check if it's not None
        module = inspect.getmodule(current_frame.f_back)
        if module is None:
            raise AttributeError(
                f"Could not get module name for tracing (name={name}, attributes={attributes})"
            )

        tracer = trace.get_tracer(module.__name__)
        attributes = attributes or {}

        with tracer.start_as_current_span(name=name, attributes=attributes):
            yield
    except AttributeError as e:
        logger.error(str(e))


def jelly_trace(func=None, *, span_name=None):
    """
    Decorator for tracing a function with Jellyfish's telemetry libraries.
    We use a convention where the decorated function's module's name is used as the tracer name.
    """

    # While name is optional, this decorator will not execute unless all parameters are provided.
    if func is None:
        return partial(jelly_trace, span_name=span_name)

    # Use the module name of the calling function as the tracer name.
    module = sys.modules[func.__module__]
    tracer = trace.get_tracer(module.__name__)

    # If the span name is not provided, use the function name.
    span_name = span_name or func.__name__

    if asyncio.iscoroutinefunction(func):

        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            otel_span = tracer.start_span(name=span_name)
            try:
                return await func(*args, **kwargs)
            finally:
                otel_span.end()

        return async_wrapper

    otel_trace = tracer.start_as_current_span(name=span_name)

    return otel_trace(func)



import logging
from typing import Any, Generator, NamedTuple, Optional

import requests

from jf_ingest import logging_helper
from jf_ingest.config import UserMetadata
from jf_ingest.utils import (
    get_jellyfish_api_base_url,
    get_jellyfish_api_token,
    retry_for_status,
)


class JiraFieldIdentifier(NamedTuple):
    """When scrubbing field information using include_fields or exclude_fields
    Jira is not always consistent in how it identifies the fields. From the
    fields API endpoint (downloaders.download_fields) you can reliable use
    the "id" field (something like "customfield_XXXXXXXX"), but when scrubbing
    changelogs on issues (via _filter_changelogs) you sometimes can only use
    the human readable name (like "Your Custom Field"). This Tuple will hold both the
    Jira ID and the human readable name.

    For premade fields in Jira, like "Description", the ID field is typically the same
    as the human readable name but with/without capitalization. Example: 'description' vs 'Description'
    """

    jira_field_id: str  # The Jira ID field. Can be premade (like 'description') or custom (like 'customfield_XXXXXX')
    jira_field_name: str  # The human readable field name. Can be premade (like 'Description') or custom like 'Your Custom Field'

    def matches_changelog_item(self, changelog_item: dict):
        if 'fieldId' in changelog_item:
            return self.jira_field_id == changelog_item['fieldId']
        elif 'field' in changelog_item:
            changelog_field_name: str = changelog_item['field']
            jira_field_name = self.jira_field_name
            # NOTE: When Agent customers want to include fields that are specific to Jira
            # (think description, status, assignee), it generally doesn't use capitalization
            # for the name value of the field. To get around this, normalize everything to
            # lowercase when dealing with. Since these all come from Jira, I am not worried
            # about capitalization collisions being an issue. This is only relevant for Jira
            # Server instances, where the fieldId is generally not included
            if field_type := changelog_item.get('fieldtype', ''):
                if field_type == 'jira':
                    changelog_field_name = changelog_field_name.lower()
                    jira_field_name = jira_field_name.lower()
            return jira_field_name == changelog_field_name
        else:
            logging_helper.log_standard_error(
                level=logging.WARNING,
                error_code=3082,
                msg_args=[changelog_item.keys()],
            )
            return False


def get_jellyfish_jira_issues_count(
    jellyfish_api_base_url: Optional[str] = None, jellyfish_api_token: Optional[str] = None
) -> int:
    """Helper function for getting the total number of issues that exist in Jellyfish

    Args:
        jellyfish_api_base_url (Optional[str]): Used as the base API to get data from jellyfish. If not provided, we will attempt to read it from the global variable.
        jellyfish_api_token (Optional[str]): Used for authenticating against Jellyfish. If not provided, we will attempt to read it from the global variable.

    Returns:
        int: The total number of Jellyfish Jira Issues that exist in this customers instance
    """
    if not jellyfish_api_base_url:
        jellyfish_api_base_url = get_jellyfish_api_base_url()
    if not jellyfish_api_token:
        jellyfish_api_token = get_jellyfish_api_token()

    base_url = jellyfish_api_base_url
    headers = {"Jellyfish-API-Token": jellyfish_api_token}

    r = retry_for_status(
        requests.get,
        f"{base_url}/endpoints/jira/issues/count",
        headers=headers,
    )
    r.raise_for_status()
    json_resp: dict = r.json()
    num_issues: int = json_resp.get('total_issues_in_jellyfish', 0)

    return num_issues


def expand_and_normalize_jira_fields(
    fields_from_jira: list[dict], field_names_or_ids: list[str]
) -> list[JiraFieldIdentifier]:
    """Expands/Translates whatever the Agent config has for include or exclude fields, and
    transforms them to be a list of JiraFieldIdentifier objects which contain both the name and the ID.

    For example, if the exclude_list has the following fields: ['Sprint'],
    which represents the human readable name for each field, and the Jira Fields API endpoint
    returns us {'id': 'customfield_0001', 'name': 'Sprint'}, this function would transform the
    exclude_list to be {jira_field_name: 'Sprint', jira_field_name: 'customfield_0001'}.
    If the exclude_list was ['customfield_0001'] would provide the same result
    NOTE: DO NOT EVERY MESS WITH CAPITALIZATION, AS THE JIRA API IS CASE SENSITIVE


    Args:
        fields_from_jira (list[dict]): The raw JSON fields returned from the field API endpoint
        field_names_or_ids (list[str]): A list of field IDs or field Names to expand and/or normalize

    Returns:
        : A list of JiraFieldIdentifier objects that contain both the field ID and the field Name
    """
    field_id_to_name_lookup = {field['id']: field['name'] for field in fields_from_jira}
    field_name_to_id_lookup = {field['name']: field['id'] for field in fields_from_jira}
    jira_field_identifiers = set()
    for provided_field in field_names_or_ids:
        field_name = field_id_to_name_lookup.get(provided_field)
        field_id = field_name_to_id_lookup.get(provided_field)
        if field_name:
            jira_field_identifiers.add(
                JiraFieldIdentifier(jira_field_id=provided_field, jira_field_name=field_name)
            )
        if field_id:
            jira_field_identifiers.add(
                JiraFieldIdentifier(jira_field_id=field_id, jira_field_name=provided_field)
            )
        # If we don't have a field_name or a field_id, then we might have a provided field
        # that exists on an issue but is not included in the returned data from the /field
        # endpoint. We've seen this for at least one client when it comes to the "parent" field
        if not (field_name or field_id):
            jira_field_identifiers.add(
                JiraFieldIdentifier(jira_field_id=provided_field, jira_field_name=provided_field)
            )

    return list(jira_field_identifiers)


def get_user_key(gdpr_active: bool) -> str:
    """Helper function for getting the name of the key that maps
    to the unique identify for a user in the Jira Instance. This
    value can be either "key" or "accountId". "accountId" is used
    by Jira Cloud, and "key" is used by Jira Server

    Args:
        gdpr_active (bool): Switch to determine what type of Jira
        Instance we're in

    Returns:
        str: Either "key" or "accountId", depending on the
        instance type
    """

    # Choose the key name based on the GDPR status
    if gdpr_active:
        return "accountId"
    else:
        return "key"


def get_user_key_from_user(user_dict: dict, gdpr_active: bool = False, **kwargs) -> str:
    """Helper function used for getting unique set of users

    Args:
        user_dict (dict): Raw User dict from JIRA API
        gdpr_active (bool, optional): Switches what key to grab, depending on if we are server or cloud. Defaults to False.

    Raises:
        KeyError: _description_

    Returns:
        _type_: Jira User Unique key (accountId or Key, depending on gdpr_active)
    """
    key_name = get_user_key(gdpr_active=gdpr_active)

    # Return a default value if one is provided, otherwise raise a KeyError
    try:
        if "default" in kwargs:
            default_value = kwargs["default"]
            kn: str = user_dict.get(key_name, default_value)
        else:
            kn = user_dict[key_name]

        return kn
    except KeyError as e:
        raise KeyError(
            f'Error extracting user data from Jira data. GDPR set to "{gdpr_active}" and expecting key name: "{key_name}" in user_dict. This is most likely an issue with how the GDPR flag is set on Jira instance. If this is a Jira Agent configuration, the agent config.yml settings may also be wrong.'
        ) from e


def get_unique_users_from_issue(issue: dict, gdpr_active: bool) -> set[UserMetadata]:
    """Extracts all the user keys (either key or accountId, depending on if it's server or cloud)
    from a jira issue. This is useful for getting all the users that are associated with an issue
    and seeing if we need to download additional users

    Args:
        issue (dict): A raw issue dictionary from the Jira API

    Returns;
        set[str]: A set of user keys (either key or accountId) that are associated with the issue
    """
    user_key = get_user_key(gdpr_active=gdpr_active)
    users: set[UserMetadata] = set()

    def _recurse_on_issue_data_for_user_key(dict_or_list: Any):
        if isinstance(dict_or_list, dict):
            if 'self' in dict_or_list and '/rest/api/2/user' in dict_or_list['self']:
                if user_key not in dict_or_list:
                    return
                users.add(
                    UserMetadata(
                        self_url=dict_or_list['self'],
                        user_key=dict_or_list[user_key],
                        email=dict_or_list.get('email'),
                        name=dict_or_list.get('name'),
                        active=dict_or_list.get('active'),
                    )
                )
                return
            for key, value in dict_or_list.items():
                if isinstance(value, (dict, list)):
                    _recurse_on_issue_data_for_user_key(value)
        if isinstance(dict_or_list, list):
            for value in dict_or_list:
                if isinstance(value, (dict, list)):
                    _recurse_on_issue_data_for_user_key(value)

    _recurse_on_issue_data_for_user_key(issue)
    return users



import contextvars
import dataclasses
import hashlib
import inspect
import io
import json
import logging
import re
import sys
import time
import traceback
from collections import namedtuple
from concurrent.futures import FIRST_COMPLETED, Future, ThreadPoolExecutor, wait
from datetime import datetime, timezone
from enum import Enum
from types import TracebackType
from typing import Any, Callable, Generator, Iterable, Optional, Type, cast

import gitlab
import requests
import urllib3
from dateutil import parser as date_parser
from jira.exceptions import JIRAError
from requests.adapters import HTTPAdapter
from requests.cookies import RequestsCookieJar, cookiejar_from_dict
from tqdm import tqdm
from urllib3 import Retry

from jf_ingest import logging_helper
from jf_ingest.telemetry.metrics import JellyCounter

RETRY_FOR_STATUS_CALL_COUNTER: Optional[JellyCounter] = None


def setup_retry_counter() -> JellyCounter:
    return JellyCounter(
        name="jf_ingest_retry_for_status",
        description="Total number of calls attempted by retry_for_status, includes retry and non retry events",
    )


logger = logging.getLogger(__name__)

# Exponent base of 4 gives more reasonable
# max timeout of 4m16 rather than 10m25s for
# 5 retries
RETRY_EXPONENT_BASE = 4
DEFAULT_HTTP_CODES_TO_RETRY_ON = (429, 500, 502, 503, 504)

# For some Jira Servers, we can get fed a 401 error that is actually
# transient and totally retryable. Although retrying on all 401s will
# slightly slowdown raising actual authentication errors, the retry
# logic here is worth it. Gavin, December 2024
JIRA_STATUSES_TO_RETRY = tuple(list(DEFAULT_HTTP_CODES_TO_RETRY_ON) + [401])
# 500 errors are ignorable, and typically mean that the board is not configured for sprints.
# We are safe to log and ignore 500 level errors
JIRA_SPRINT_ERRORS_TO_RETRY = tuple(
    [status_code for status_code in JIRA_STATUSES_TO_RETRY if status_code != 500]
)
GITLAB_STATUSES_TO_RETRY = tuple(
    list(DEFAULT_HTTP_CODES_TO_RETRY_ON) + [522]
)  # We can get these transient Cloudflare Timeouts, attempt a retry

# There appears to be intermittent 404s that get thrown for versions and components.
# When we retry, the 404 goes away.
# NOTE - elheureux 2024-11-01: Adding 500 to the list of codes to retry on, as we are seeing
# intermittent 500 internal server errors when trying to access project components
# Additional context: OJ-39786
PROJECT_HTTP_CODES_TO_RETRY_ON = tuple(list(JIRA_STATUSES_TO_RETRY) + [404])

DEFAULT_EXCEPTIONS_TO_RETRY_ON = (
    requests.exceptions.ReadTimeout,
    requests.exceptions.ConnectionError,
    requests.exceptions.ConnectTimeout,
    requests.exceptions.ChunkedEncodingError,
    urllib3.exceptions.ConnectTimeoutError,
    urllib3.exceptions.ProtocolError,
)


class GitLabGidObjectMapping(Enum):
    PROJECT = 'Project'
    MERGE_REQUEST = 'MergeRequest'
    USER = 'User'
    GROUP = 'Group'


class RetryLimitExceeded(Exception):
    pass


class StrDefaultEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return str(o)


class ReauthSession(requests.Session):
    """
    A requests session that will re-authenticate on 401s
    """

    # Redefining this for mypy to get type. This is identical to `requests.Session`.
    cookies: RequestsCookieJar = cookiejar_from_dict({})

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def request(self, method, url, **kwargs) -> requests.Response:
        """
        Run a request, and if we get a 401, re-authenticate and try again.
        @:param: method: The HTTP method to use
        @:param: url: The URL to request
        @:param: kwargs: Any additional kwargs to pass to the request
        (a reauth session is usually instantiated with just ReauthSession(**kwargs))
        @:return: The response object from calling request()
        """
        # If we get HTTP 401, re-authenticate and try again
        response = super().request(method, url, **kwargs)
        if response.status_code == 401:
            # Use print instead of logger.log, as URL could be considered sensitive data
            logger.warn(f"Received 401 for the request [{method}] {url} - resetting client session")

            # Clear cookies and re-auth
            self.cookies.clear()
            response = super().request(method, url, **kwargs)
            self.cookies = response.cookies
        return response


def get_attribute(object, property, default=None):
    """
    Obtain a class attribute safely
    """
    try:
        value = getattr(object, property)
        return value if value else default
    except AttributeError:
        return default


def retry_session(**kwargs) -> requests.Session:
    """
    Obtains a requests session with retry settings.
    :return: session: Session
    """

    session = ReauthSession(**kwargs)

    retries = 3
    backoff_factor = 0.5
    status_forcelist = (500, 502, 503, 504, 104)

    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )

    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)

    return session


def get_wait_time(e: Optional[Exception], retries: int) -> int:
    """
    This function attempts to standardize determination of a wait time on a retryable failure.
    If the exception's response included a Retry-After header, respect it.
    If it does not, we do an exponential backoff - 1s, 4s, 16s, 64s, 256s.

    A possible future addition would be to add a jitter factor.
    This is a fairly standard practice but not clearly required for our situation.
    """
    # getattr with a default works on _any_ object, even None.
    # We expect that almost always e will be a JIRAError or a RequestException, so we will have a
    # response and it will have headers.
    # So I'm choosing to use the getattr call to handle the valid but infrequent possibility
    # that it may not (None or another Exception type that doesn't have a response), rather tha
    # preemptively checking.
    response = getattr(e, "response", None)
    headers = getattr(response, "headers", {})
    retry_after = headers.get("Retry-After")

    # Normalize retry after if it is a string
    if isinstance(retry_after, str) and retry_after.isnumeric():
        retry_after = int(retry_after)
    # Don't do anything if we have a valid int for retry after
    elif isinstance(retry_after, int):
        pass
    else:
        # Null out any invalid retry after values
        retry_after = None

    if retry_after:
        return retry_after
    else:
        return int(RETRY_EXPONENT_BASE**retries)


def called_from_module(module: str) -> bool:
    """
    Helper function to determine if the current function is being called from a specific module
    :param module: The module name to check against
    :return: True if the current function is being called from the specified module
    """
    stack = inspect.stack()
    for frame in stack:
        module_name = frame.frame.f_globals.get('__name__')
        if module_name and module in module_name:
            return True
    return False


def get_caller_module() -> str:
    """hellper function to determine the module that called the current function

    Returns:
        str: The Module name that called the current function
    """
    stack = inspect.stack()
    for frame in stack:
        module_name: str = frame.frame.f_globals.get('__name__', '')
        if module_name and 'jf_ingest' in module_name:
            return module_name

    return RequestModuleSource.UNKNOWN.value


class RequestModuleSource(Enum):
    """Enum to help us determine where requests are coming from
    when using the retry_for_status helper function
    """

    JIRA = 'JIRA'
    AZURE_DEVOPS = 'AZURE_DEVOPS'
    GITLAB = 'GITLAB'
    GITHUB = 'GITHUB'
    UNKNOWN = 'UNKNOWN'


def get_request_source() -> RequestModuleSource:
    """Determines the module source of the caller

    Returns:
        RequestModuleSource: Returns a RequestModuleSource enum type
    """
    if called_from_module('jf_ingest.jf_jira'):
        return RequestModuleSource.JIRA
    elif called_from_module('jf_ingest.jf_git.clients.azure_devops'):
        return RequestModuleSource.AZURE_DEVOPS
    elif called_from_module('jf_ingest.jf_git.clients.gitlab'):
        return RequestModuleSource.GITLAB
    elif called_from_module('jf_ingest.jf_git.clients.github'):
        return RequestModuleSource.GITHUB
    else:
        return RequestModuleSource.UNKNOWN


# Specific Jira Retry headers to try to grab
JiraRetryHeaders = namedtuple(
    'JiraRetryHeaders',
    (
        'X_RateLimit_Limit',
        'X_RateLimit_Remaining',
        'RateLimit_Reason',
        'X_RateLimit_NearLimit',
        'Beta_Retry_After',
        'X_Beta_RateLimit_NearLimit',
        'X_Beta_RateLimit_Reason',
        'X_Beta_RateLimit_Reset',
    ),
)


def get_jira_headers(resp: requests.Response) -> JiraRetryHeaders:
    """Generates a named tuple JiraRetryHeaders object from a requests.Response object,
    containing the Jira Rate Limit Headers. It is likely that Jira didn't supply us these
    headers, but it's worth attempting to grab.
    Headers are stated to be present in the Jira API documentation, here:
    https://developer.atlassian.com/cloud/jira/platform/rate-limiting/#rate-limit-related-headers

    Args:
        resp (requests.Response): A requests.Reponse object returned from any API call, not gaurenteed
        to have been a call for Jira

    Returns:
        JiraRetryHeaders: A named tuple object representing the headers present in the Jira API response
    """
    return JiraRetryHeaders(
        X_RateLimit_Limit=resp.headers.get('X-RateLimit-Limit'),
        X_RateLimit_Remaining=resp.headers.get('X-RateLimit-Remaining'),
        RateLimit_Reason=resp.headers.get('RateLimit-Reason'),
        X_RateLimit_NearLimit=resp.headers.get('X-RateLimit-NearLimit'),
        Beta_Retry_After=resp.headers.get('Beta-Retry-After'),
        X_Beta_RateLimit_NearLimit=resp.headers.get('X-Beta-RateLimit-Reset'),
        X_Beta_RateLimit_Reason=resp.headers.get('X-Beta-RateLimit-Reason'),
        X_Beta_RateLimit_Reset=resp.headers.get('X-Beta-RateLimit-Reset'),
    )


def retry_for_status(
    f: Callable[..., Any],
    *args,
    max_retries_for_retry_for_status: int = 5,
    statuses_to_retry: Iterable[int] = DEFAULT_HTTP_CODES_TO_RETRY_ON,
    exceptions_to_retry: Iterable[Type[Exception]] = DEFAULT_EXCEPTIONS_TO_RETRY_ON,
    retry_on_any_exception: bool = False,
    **kwargs,
) -> Any:
    """
    This function allows for us to retry for a variety of retryable errors from Jira.
    There are much more elegant ways of accomplishing this, but this is a quick and
    reasonable approach to doing so.

    Default statuses are 429 (rate limit exceeded) and 5xx errors (server errors).

    Note:
        - max_retries=5 will give us a maximum wait time of 4m16s.
    """
    global RETRY_FOR_STATUS_CALL_COUNTER
    if not RETRY_FOR_STATUS_CALL_COUNTER:
        RETRY_FOR_STATUS_CALL_COUNTER = setup_retry_counter()

    request_source = get_request_source().value
    function_name = getattr(f, '__name__', 'Unknown_Function')
    if request_source == RequestModuleSource.UNKNOWN.value:
        request_source = get_caller_module()

    for retry in range(max_retries_for_retry_for_status + 1):

        jelly_counter_attributes: dict[str, Any] = {
            "function_name": function_name,
            "max_retries": max_retries_for_retry_for_status,
            "request_source": request_source,
            "retry_attempt": retry,
            "company_slug": get_jellyfish_company_slug() or 'unknown',
        }

        # Init None Return value
        return_val = None
        try:
            return_val = f(*args, **kwargs)
            # If we pass in a session.get or session.post call to this function,
            # an error is not immediately raised because you need to call this
            # raise_for_status class method on the Response object first
            if isinstance(return_val, requests.Response):
                jelly_counter_attributes['status_code'] = return_val.status_code
                # Raise the Error
                return_val.raise_for_status()

                # If error wasn't raised, log the specific jira headers (if present)
                jira_retry_headers = get_jira_headers(return_val)
                if jira_retry_headers.X_Beta_RateLimit_NearLimit != None:
                    jelly_counter_attributes['near_beta_jira_limit'] = (
                        jira_retry_headers.X_Beta_RateLimit_NearLimit
                    )

                if jira_retry_headers.Beta_Retry_After != None:
                    jelly_counter_attributes['over_beta_jira_limit'] = bool(
                        jira_retry_headers.Beta_Retry_After
                    )

                if jira_retry_headers.X_RateLimit_NearLimit != None:
                    jelly_counter_attributes['near_current_jira_limit'] = bool(
                        jira_retry_headers.X_RateLimit_NearLimit
                    )

            jelly_counter_attributes['is_retry'] = False
            RETRY_FOR_STATUS_CALL_COUNTER.add(1, attributes=jelly_counter_attributes)
            return return_val
        except Exception as e:
            if retry >= max_retries_for_retry_for_status:
                logging_helper.send_to_agent_log_file(
                    f'Retry Limit reached when running function {function_name} from {request_source=}. Raising error:',
                    level=logging.ERROR,
                )
                # Raise any non-429 related errors
                logging_helper.send_to_agent_log_file(traceback.format_exc(), level=logging.ERROR)
                raise RetryLimitExceeded(e)
            is_handled_exception = any(
                isinstance(e, exception) for exception in exceptions_to_retry
            )

            # Get the status code from the exception, if it exists
            if isinstance(e, requests.exceptions.HTTPError) and hasattr(e, 'response'):
                status_code = getattr(e.response, 'status_code', None)
            # Special case for Gitlab errors, where status code is hidden as response code
            elif isinstance(e, gitlab.exceptions.GitlabError):
                status_code = getattr(e, 'response_code', None)
            else:
                status_code = getattr(e, "status_code", None)

            # Ignoring for mypy, since statuses_to_retry and status_code are mixed.
            if not (is_handled_exception or status_code in statuses_to_retry) and not retry_on_any_exception:  # type: ignore[operator]
                raise e

            # Otherwise, retry

            # Get Wait Time
            wait_time = get_wait_time(e, retries=retry)

            # Log Error
            error_code = 3071
            exception_name = getattr(e.__class__, '__name__', 'Exception')
            msg_args = [
                (
                    f"err={exception_name},module={request_source},status={status_code}"
                    if status_code
                    else f"err={exception_name},module={request_source}"
                ),
                retry + 1,
                max_retries_for_retry_for_status,
                wait_time,
            ]
            logging_helper.log_standard_error(
                logging.WARNING,
                msg_args=msg_args,
                error_code=error_code,
            )

            # Increment Counter
            if status_code:
                jelly_counter_attributes['status_code'] = status_code
            jelly_counter_attributes['exception'] = exception_name
            jelly_counter_attributes['is_retry'] = True
            RETRY_FOR_STATUS_CALL_COUNTER.add(
                1,
                attributes=jelly_counter_attributes,
            )

            # Sleep
            time.sleep(wait_time)
            continue


def test_jira_or_git_object_access(
    f: Callable[..., Any],
    *args,
    is_sprint: bool = False,
    return_objs: bool = True,
    return_attr: Optional[str] = None,
    **kwargs,
) -> tuple[bool, list[Any]]:
    """
    Determines whether we can access a particular class of objects within jira or git. In general, this returns True
    if no errors are thrown and False otherwise. Sprints are a special case. Boards may not support sprints,
    the board could be misconfigured, etc. If we are looking for a sprint and get a 400, 404, or 500 error
    response, assume that everything is ok and we're not missing any access.

    Args:
        f (Callable[..., Any]): JIRA.<function>
        is_sprint (bool, optional): whether we are looking for sprint data. Defaults to False. Only applicable to Jira
        return_obj (bool, optional): whether we want to return some data from the accessed objects. Returns an empty list if False. Defaults to True.
        return_attr (Any): attribute to return in a list from the accessed objects.

    Returns:
        bool: whether access to the specified object type is available.
    """

    return_attr = return_attr or 'name'

    def _get_return_list(objs: Iterable) -> list:
        objs = list(objs)  # put generator items into memory if we're dealing with git
        if not return_objs or not objs:
            return []
        if isinstance(objs[0], dict):
            return [d.get(return_attr) for d in objs]
        if not hasattr(objs[0], return_attr):
            return []
        return [getattr(obj, return_attr) for obj in objs]

    try:
        objs = retry_for_status(f, *args, **kwargs)
        return True, _get_return_list(objs)
    except JIRAError as e:
        if is_sprint:
            return e.status_code in [400, 404, 500], []
        logger.debug(
            f'Jira Error ({e.status_code}) encountered when attempting to hit function {f.__name__}. Error: {e}'
        )
        return False, []
    except Exception as e:
        logger.debug(f'Error encountered when attempting to hit function {f.__name__}. Error: {e}')
        return False, []


def batch_iterable(iterable: Iterable, batch_size: int) -> Generator[list[Any], None, None]:
    """Helper function used for batching a given iterable into equally sized batches

    Args:
        iterable (Iterable): An iterable you want to split into batches
        batch_size (int): The size of the batches you want

    Yields:
        Generator[list[Any], None, None]: This generator yields a list of equal size batches, plus a potential final batch that is less than the batch_size arg
    """
    chunk = []
    i = 0
    for item in iterable:
        chunk.append(item)
        i += 1
        if i == batch_size:
            yield chunk
            chunk = []
            i = 0

    if chunk:
        yield chunk


def get_object_bytes_size(obj, seen=None):
    """Recursively finds size of objects"""
    size = 0
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_object_bytes_size(v, seen) for v in obj.values()])
        size += sum([get_object_bytes_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += get_object_bytes_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_object_bytes_size(i, seen) for i in obj])
    else:
        size += sys.getsizeof(obj)
    return size


def batch_iterable_by_bytes_size(
    iterable: Iterable, batch_byte_size: int
) -> Generator[list[Any], None, None]:
    """Helper function that batches an iterable by it's total size in bytes

    Args:
        iterable (Iterable): An iterable type object that you want batched
        batch_byte_size (int): The total size of each batch size, in bytes

    Yields:
        Generator[list[Any], None, None]: A generator that returns roughly equal sized batches of data (in byte size)
    """
    chunk = []
    current_chunk_size = 0
    for item in iterable:
        chunk.append(item)
        current_chunk_size += get_object_bytes_size(item)

        if current_chunk_size >= batch_byte_size:
            yield chunk
            # Reset batch information
            chunk = []
            current_chunk_size = 0

    # Write the final batch, if applicable
    if chunk:
        yield chunk


def format_date_to_jql(_datetime: datetime) -> str:
    """Formats a python datetime to a str in this format: YYYY-MM-DD, which is JQL friendly

    Args:
        datetime (datetime): A valid Python Datetime

    Returns:
        str: Returns a formatted datetime str in this format (with padded 0s, if needed): YYYY-MM-DD
    """
    # Special case: if we've got date.min, always return year 1, month 1, day 1.
    # This value should really never be provided, but for legacy reasons we used
    # to mark issues needing redownload as having an updated value as datetime.min
    if _datetime == datetime.min:
        return f'"0001-01-01"'

    return f'"{_datetime.year:04}-{_datetime.month:02}-{_datetime.day:02}"'


def format_datetime_to_ingest_timestamp(datetime: datetime):
    return datetime.strftime("%Y%m%d_%H%M%S")


class TqdmToLogger(io.StringIO):
    """
    Output stream for TQDM which will output to logger module instead of
    the StdOut.
    """

    logger = None
    level = None
    buf = ''

    def __init__(self, logger, level=None):
        super(TqdmToLogger, self).__init__()
        self.logger = logger
        self.level = level or logging.INFO

    def write(self, buf):
        self.buf = buf.strip('\r\n\t ')

    def flush(self):
        self.logger.log(self.level, self.buf)


def tqdm_to_logger(*args, **kwargs) -> tqdm:
    """HACK helper function to wrap up the TQDM progress bar to extend it's functionality.
    If you use this helper function in place of the normal tqdm() call it will behave differently
    in Agent and in Direct Connect. In Agent, TQDM will act as if you called it directly.
    If you call this function in a DIRECT_CONNECT context, it will print each TQDM progress
    update to our logger wrapper (at the INFO level), which will get caught up in Prefect.

    Returns:
        tqdm: A TQDM object
    """
    from jf_ingest.config import IngestionType

    ingestion_type = get_ingestion_type()
    if ingestion_type == IngestionType.DIRECT_CONNECT:
        tqdm_out = TqdmToLogger(logger, level=logging.INFO)
        # Log progress bar every 10 minutes (minimum) in prefect
        mininterval = kwargs.get('mininterval', 60 * 10)
        return tqdm(file=tqdm_out, mininterval=mininterval, ascii=False, *args, **kwargs)
    return tqdm(*args, **kwargs)


class ThreadPoolExecutorWithLogging(ThreadPoolExecutor):
    def submit(self, fn, *args, **kwargs):
        ctx = contextvars.copy_context()

        def _wrapped_fn(*args_inner, **kwargs_inner):
            return ctx.run(fn, *args_inner, **kwargs_inner)

        future = super().submit(_wrapped_fn, *args, **kwargs)
        return future


class ThreadPoolWithTqdm(ThreadPoolExecutorWithLogging):
    """THIS CLASS MUST BE USED AS A CONTEXT MANAGER ONLY!!!!

    This is a custom class that extends the ThreadPoolExecutor class,
    and combines it with some TQDM (progress bar) functionality. This
    should help reduce the number of repeated code around jf_ingest

    Yes, I know there is a concurrency extension for TQDM (tqdm.contrib.concurrent)
    BUT this library only has .map style functions, and it does NOT have a submit style
    function. There are instances where using submit is preferred, and often it lends itself
    to simpler code. That is what this library is for

    Args:
        ThreadPoolExecutor (ThreadPoolExecutor): The parent class that this extends

    Returns:
        ThreadPoolWithTqdm: ThreadPoolWithTqdm
    """

    desc: str
    total: int
    futures: set[Future]
    raise_exceptions: bool
    prog_bar: tqdm
    hide_tqdm: bool

    def __init__(
        self,
        desc: Optional[str] = None,
        total: int = 0,
        max_workers: int | None = None,
        raise_exceptions: bool = False,
        thread_name_prefix: str = "",
        initializer: Callable[..., object] | None = None,
        initargs: tuple[Any, ...] = (...,),
        hide_tqdm: Optional[bool] = False,
    ) -> None:
        """Custom Constructor, to allow us to set the progress bar values

        Args:
            desc (str, optional): The description field on the TQDM progress bar. Defaults to None.
            total (int, optional): The total value to set on the TQDM progress bar. Defaults to 0.
            max_workers: The maximum number of threads that can be used to execute the given calls.
            thread_name_prefix: An optional name prefix to give our threads.
            initializer: A callable used to initialize worker threads.
            initargs: A tuple of arguments to pass to the initializer.

        """

        self.desc = desc or ''
        self.total = total
        self.futures = set()
        self.raise_exceptions = raise_exceptions
        self.hide_tqdm = bool(hide_tqdm)
        super().__init__(max_workers, thread_name_prefix, initializer, initargs)

    def __enter__(self):
        """Override the __enter__ function to instantiate a progress bar (TQDM)
        when using this as a context manager.
        """
        self.prog_bar = tqdm_to_logger(desc=self.desc, total=self.total, disable=self.hide_tqdm)
        return super().__enter__()

    def submit(self, fn: Callable, *args, **kwargs) -> Future:  # type: ignore[override]
        future: Future = super().submit(fn, *args, **kwargs)
        self.futures.add(future)
        return future

    def get_results(self) -> Generator[Any, None, None]:
        while self.futures:
            done_futures, _ = wait(self.futures, return_when=FIRST_COMPLETED)
            for done_future in done_futures:
                # Update the progress bar each time we get a future back
                try:
                    self.update_progress_bar(done_future)
                    yield done_future.result()
                except Exception as e:
                    logging_helper.send_to_agent_log_file(
                        f'Exception encountered in ThreadPoolExecutor: {e}',
                        level=logging.ERROR,
                        exc_info=True,
                    )
                    if self.raise_exceptions:
                        raise
                finally:
                    # Remove the future from the pool, for memory performance management
                    self.futures.remove(done_future)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        """Override the __exit__ function to tear down a progress bar (TQDM)
        when using this as a context manager.
        Progress bar .close() call MUST BE AFTER THE THREADPOOLEXECUTOR TEARDOWN.

        Also raises any exceptions we encountered
        """

        # Always empty the thread pool before exiting, so that the progress bar increments
        # even if nobody ever called the get_results function
        if self.futures:
            for _ in self.get_results():
                # Iterate through the get_results generate to force progress bar load
                pass
        retval = super().__exit__(exc_type, exc_val, exc_tb)

        # Progress bar must be close AFTER the parent __exit__ call,
        # because the parent __exit__ call is blocking
        self.prog_bar.close()

        return retval

    def update_progress_bar(self, future: Future):
        """A custom callback function that iterates on the progress bar by
        default. Does some guessing to see how much we should updated the
        progress bar by

        Args:
            future (Future): A ThreadPoolExecutor Future object
        """
        if future.exception():
            return

        result = future.result()

        with tqdm.get_lock():
            if isinstance(result, Iterable):
                self.prog_bar.update(len(list(result)))
            else:
                self.prog_bar.update(1)


def get_s3_base_path(company_slug: str, timestamp: str):
    return f"{company_slug}/{timestamp}"


class RewriteJiraSessionHeaders(object):
    """
    This context manager will temporarily rewrite the headers of the JIRA session to
    only use the Accept: application/json header, no other additional value. This gets around
    the incident where resolutions endpoint failed with HTTP 406 errors
    https://jelly-ai.atlassian.net/browse/OJ-31563
    """

    saved_old_accept_headers = ""
    jira_connection = None

    def __init__(self, jira_connection):
        self.jira_connection = jira_connection

    def __enter__(self):
        self.old_headers = self.jira_connection._session.headers
        self.saved_old_accept_headers = self.jira_connection._session.headers["Accept"]
        self.jira_connection._session.headers["Accept"] = "Application/json"

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.jira_connection._session.headers["Accept"] = self.saved_old_accept_headers


def chunk_iterator_in_lists(n, iterable) -> list[list]:
    """
    :param n: The size of the chunks
    :param iterable: The iterable to chunk
    :return: A list of lists where each inner list is of size n with the final inner list having size len(iterable) % n
    if n == 0 then return [[], [<all elements>]] as it produces a chunk of size 0, and then all remainders
    """
    parent_list = []
    counter = 0
    while counter + n <= len(iterable):
        parent_list.append(iterable[counter : counter + n])
        counter += n
    if counter < len(iterable):
        parent_list.append(iterable[counter : len(iterable)])
    return parent_list


def hash_filename(filepath: str):
    """
    Helper function for files. Hashes a filepath, tokenized by '/'.
    :param filepath: Input filepath.
    """

    def _compute_hash(string):
        return hashlib.md5(string.encode('utf-8'), usedforsecurity=False).hexdigest()[:12]

    extension = filepath.split(".")[-1]
    paths_list = filepath.split("/")

    # Check if a file extension (.) exists
    if len(extension) == len(filepath):
        return _compute_hash(filepath)
    return "/".join([_compute_hash(path) for path in paths_list]) + f".{extension}"


def set_ingestion_type(ingestion_type):
    """Helper function used to setting the global INGESTION_TYPE variable in this module

    Args:
        ingestion_type (_type_): The Ingestion Type we are doing (Agent or Direct Connect). Accepts the ENUM from the config.py file!
    """
    global INGESTION_TYPE
    INGESTION_TYPE = ingestion_type
    logger.debug(f'Set global value INGESTION_TYPE to {ingestion_type}')


def get_ingestion_type():
    from jf_ingest.config import IngestionType

    if 'INGESTION_TYPE' in globals():
        global INGESTION_TYPE
        return INGESTION_TYPE
    else:
        default_ingestion_type = IngestionType.AGENT
        logging_helper.send_to_agent_log_file(
            f'Ingestion type not set. This may affect the logging of this run. Defaulting to ingestion type of {default_ingestion_type}',
            level=logging.WARNING,
        )
        return default_ingestion_type


# Defined globals outside of functions for mypy
COMPANY_SLUG = ''
global JELLYFISH_API_BASE_URL
global JELLYFISH_API_TOKEN


def set_jellyfish_company_slug(company_slug: str):
    """Helper function used to setting the global COMPANY_SLUG variable in this module

    Args:
        jellyfish_api_base_url (str): The company_slug
    """
    global COMPANY_SLUG
    COMPANY_SLUG = company_slug
    logger.debug(f'Setting company slug to {company_slug}')


def get_jellyfish_company_slug():
    if 'COMPANY_SLUG' in globals():
        global COMPANY_SLUG
        return COMPANY_SLUG
    else:
        return None


def set_jellyfish_api_base_url(jellyfish_api_base_url: str):
    """Helper function used to setting the global JELLYFISH_API_BASE_URL variable in this module

    Args:
        jellyfish_api_base_url (str): The Jellyfish API Base URL
    """
    JELLYFISH_API_BASE_URL = jellyfish_api_base_url
    logger.debug('Setting Jellyfish API Base URL.')


def get_jellyfish_api_base_url():
    if 'JELLYFISH_API_BASE_URL' in globals():
        global JELLYFISH_API_BASE_URL
        return JELLYFISH_API_BASE_URL
    else:
        return None


def set_jellyfish_api_token(jellyfish_api_token: str):
    """Helper function used to setting the global JELLYFISH_API_TOKEN variable in this module

    Args:
        jellyfish_api_token (str): The Jellyfish API token
    """
    JELLYFISH_API_TOKEN = jellyfish_api_token
    logger.debug('Setting Jellyfish API token')


def get_jellyfish_api_token():
    if 'JELLYFISH_API_TOKEN' in globals():
        global JELLYFISH_API_TOKEN
        return JELLYFISH_API_TOKEN
    else:
        return None


def init_jf_ingest_run(ingestion_config: Any) -> None:
    from jf_ingest.config import IngestionConfig

    # Cast for mypy
    cast(IngestionConfig, ingestion_config)
    # Set Jellyfish API variables
    if not get_jellyfish_api_base_url():
        set_jellyfish_api_base_url(ingestion_config.jellyfish_api_base)
    if not get_jellyfish_api_token():
        set_jellyfish_api_token(ingestion_config.jellyfish_api_token)
    if not get_jellyfish_company_slug():
        set_jellyfish_company_slug(ingestion_config.company_slug)
    # Set Ingestion Type (has to do with logging)
    set_ingestion_type(ingestion_config.ingest_type)


def parse_gitlab_date(date):
    if date is None:
        return None

    return date_parser.parse(date)


def parse_gitlab_api_version(api_version: str) -> str:
    """Helper function
    Strips everything except integers and periods to get the semver of the provided gitlab version.

    For example, one version we've seen is 17.9.0-pre, but we're only interested in the major and minor version.

    Examples:
    17.9.0-pre -> 17.9.0
    17.9.0 -> 17.9.0
    """
    return re.sub(r'[^\d\.]', '', api_version)


def get_id_from_gid(gitlab_gid: str, object_name: str) -> str:
    """Helper function.
    The Gitlab GQL returns many IDs with this weird GID format.
    All we care about is the number trailing at the end.
    Objects are capitalized (Group, Project, MergeRequest, User, etc...)
    Gitlab Format: gid://gitlab/{object_name}/{ID_NUMBER}
    """
    return gitlab_gid.split(f'gid://gitlab/{object_name}/')[1]


def normalize_datetime(dt: Any, tz: timezone) -> Any:
    if type(dt) is datetime:
        if dt.year == 1:
            # HACK: If year is 1 than this is the default datetime.min value,
            # so it's not going to make a difference if we adjust it or not.
            # If you do attempt to adjust it you will get an overflow error.
            # To avoid all this and to keep things simple, don't bother adjusting it.
            return dt
        return dt.astimezone(tz)
    else:
        # Sometimes objects are passed in as Dates and not Datetimes,
        # so we can't adjust the timezone on them
        return dt



import json
from contextlib import contextmanager
from typing import Optional
from unittest.mock import patch

import requests_mock

from jf_ingest.jf_jira.downloaders import download_boards_and_sprints
from tests.jf_jira.utils import (
    _register_jira_uri,
    get_fixture_file_data,
    get_jira_mock_connection,
)

# NOTE: Boards and Sprints DO NOT hit the normal API, they hit the /rest/agile/1.0/ endpoint


# Dummy data for testing
PROJECT_IDS = {'12','51','1001'}
PROJECT_ID_TO_BOARDS_MAP = {
    '12': [1, 2],
    '51': [3],
    '1001': [],
}

def _mock_board_or_sprint_endpoint(
    m: requests_mock.Mocker,
    fixture_file_path: str,
    is_sprint_endpoint: bool,
    board_id: Optional[int] = None,
    project_ids: Optional[list[int]] = None,
):
    data_str = get_fixture_file_data(f"api_responses/{fixture_file_path}")
    data = json.loads(data_str)
    values = data["values"]
    total = data["total"]
    max_results = data["maxResults"]

    if total != len(values):
        # Fact check that we have good fixture data
        raise Exception(
            f"Inputted data error! Fixture {fixture_file_path} has a reported total size of {total}, but there are {len(values)} in the values payload!"
        )

    terminator_payload_str = get_fixture_file_data("api_responses/boards_empty_agile_payload.json")

    if is_sprint_endpoint:
        if board_id == None:
            raise Exception(
                f"When using the is_sprint_endpoint endpoint, board_id is a required argument!"
            )
        endpoint_base = f"board/{board_id}/sprint"
    else:
        endpoint_base = "board"
    _register_jira_uri(
        m,
        endpoint=f"{endpoint_base}?maxResults={max_results}",
        return_value=data_str,
        use_agile_endpoint=True,
    )
    _register_jira_uri(
        m,
        endpoint=f"{endpoint_base}?maxResults={max_results}&startAt={total}",
        return_value=terminator_payload_str,
        use_agile_endpoint=True,
    )
    for project_id in PROJECT_IDS:
        data['values'] = [
            board for board in values if board["id"] in PROJECT_ID_TO_BOARDS_MAP[project_id]
        ]
        _register_jira_uri(
            m,
            endpoint=f"{endpoint_base}?projectKeyOrId={project_id}&maxResults={max_results}",
            return_value=json.dumps(data),
            use_agile_endpoint=True,
        )
        _register_jira_uri(
            m,
            endpoint=f"{endpoint_base}?projectKeyOrId={project_id}&maxResults={max_results}&startAt={len(data['values'])}",
            return_value=terminator_payload_str,
            use_agile_endpoint=True,
        )

    return values


def _mark_board_as_does_not_support_sprints(
    m: requests_mock.Mocker, board_id: int, status_code: int
):
    sprints_not_supported_payload = get_fixture_file_data(
        f"api_responses/boards_sprints_not_supported.json"
    )

    _register_jira_uri(
        m,
        endpoint=f"board/{board_id}/sprint?maxResults=50",
        return_value=sprints_not_supported_payload,
        use_agile_endpoint=True,
        status_code=status_code,
    )


@contextmanager
def _mock_boards_and_sprints(
    jira_status_code_for_board_does_not_support_sprints: int = 400,
):
    with requests_mock.Mocker() as m:
        boards_from_api = _mock_board_or_sprint_endpoint(
            m, fixture_file_path="boards.json", is_sprint_endpoint=False
        )
        boards_sprints_for_board_1 = _mock_board_or_sprint_endpoint(
            m,
            fixture_file_path="boards_sprints_for_board_1.json",
            is_sprint_endpoint=True,
            board_id=1,
        )
        _mark_board_as_does_not_support_sprints(
            m,
            board_id=2,
            status_code=jira_status_code_for_board_does_not_support_sprints,
        )
        _mark_board_as_does_not_support_sprints(
            m,
            board_id=3,
            status_code=jira_status_code_for_board_does_not_support_sprints,
        )
        yield (boards_from_api, boards_sprints_for_board_1)


def test_download_boards_and_not_sprints():
    with _mock_boards_and_sprints() as board_and_sprint_tuple:
        boards_from_api = board_and_sprint_tuple[0]

        boards, sprints, links = download_boards_and_sprints(
            get_jira_mock_connection(), download_sprints=False
        )

        # Assert that we are properly parsing and getting all board data
        for board in boards:
            assert board in boards_from_api

        # Assert we are not getting sprint data, or links data
        assert sprints == []
        assert links == []


def test_download_board_with_sprints():
    with _mock_boards_and_sprints() as board_and_sprint_tuple:
        boards_from_api = board_and_sprint_tuple[0]
        boards_sprints_for_board_1 = board_and_sprint_tuple[1]

        # MOCK SPRINT DATA FOR BOARD
        boards, sprints, links = download_boards_and_sprints(
            get_jira_mock_connection(), download_sprints=True
        )

        assert boards_from_api == boards
        # Assert we are not getting sprint data, or links data
        assert sprints == boards_sprints_for_board_1
        links_from_api = [
            {
                "board_id": 1,
                "sprint_ids": [sprint["id"] for sprint in boards_sprints_for_board_1],
            },
            {"board_id": 2, "sprint_ids": []},
            {"board_id": 3, "sprint_ids": []},
        ]

        for link in links:
            assert link in links_from_api

def test_download_board_with_sprints_jira_throws_404():
    with _mock_boards_and_sprints(
        jira_status_code_for_board_does_not_support_sprints=404
    ) as board_and_sprint_tuple:
        boards_from_api = board_and_sprint_tuple[0]
        boards_sprints_for_board_1 = board_and_sprint_tuple[1]

        # MOCK SPRINT DATA FOR BOARD
        boards, sprints, links = download_boards_and_sprints(
            get_jira_mock_connection(), download_sprints=True
        )

        assert boards_from_api == boards
        # Assert we are not getting sprint data, or links data
        assert sprints == boards_sprints_for_board_1
        links_from_api = [
            {
                "board_id": 1,
                "sprint_ids": [sprint["id"] for sprint in boards_sprints_for_board_1],
            },
            {"board_id": 2, "sprint_ids": []},
            {"board_id": 3, "sprint_ids": []},
        ]

        for link in links:
            assert link in links_from_api


def test_download_board_with_sprints_jira_throws_500():
    with _mock_boards_and_sprints(
        jira_status_code_for_board_does_not_support_sprints=500
    ) as board_and_sprint_tuple, patch('jf_ingest.utils.time.sleep', return_value=0):
        boards_from_api = board_and_sprint_tuple[0]
        boards_sprints_for_board_1 = board_and_sprint_tuple[1]

        # MOCK SPRINT DATA FOR BOARD
        boards, sprints, links = download_boards_and_sprints(
            get_jira_mock_connection(), download_sprints=True
        )

        assert boards_from_api == boards
        # Assert we are not getting sprint data, or links data
        assert sprints == boards_sprints_for_board_1
        links_from_api = [
            {
                "board_id": 1,
                "sprint_ids": [sprint["id"] for sprint in boards_sprints_for_board_1],
            },
            {"board_id": 2, "sprint_ids": []},
            {"board_id": 3, "sprint_ids": []},
        ]

        for link in links:
            assert link in links_from_api


def test_download_board_with_project_filtering_smoke_test():
    with _mock_boards_and_sprints() as board_and_sprint_tuple, patch('jf_ingest.utils.time.sleep', return_value=0):
        boards_from_api = board_and_sprint_tuple[0]
        boards_sprints_for_board_1 = board_and_sprint_tuple[1]

        # MOCK SPRINT DATA FOR BOARD
        boards, sprints, links = download_boards_and_sprints(
            get_jira_mock_connection(), download_sprints=True, filter_boards_by_projects=PROJECT_IDS
        )

        boards.sort(key=lambda x: x["id"])
        boards_from_api.sort(key=lambda x: x["id"])
        assert len(boards_from_api) == len(boards)
        assert boards_from_api == boards
        # Assert we are not getting sprint data, or links data
        assert sprints == boards_sprints_for_board_1
        links_from_api = [
            {
                "board_id": 1,
                "sprint_ids": [sprint["id"] for sprint in boards_sprints_for_board_1],
            },
            {"board_id": 2, "sprint_ids": []},
            {"board_id": 3, "sprint_ids": []},
        ]

        for link in links:
            assert link in links_from_api

def test_download_board_with_project_filtering_by_projects():
    with _mock_boards_and_sprints() as board_and_sprint_tuple, patch('jf_ingest.utils.time.sleep', return_value=0):

        def _test_filter_by_board(project_id_set: set[str]) -> list[dict]:
            boards, _, _ = download_boards_and_sprints(
                get_jira_mock_connection(), download_sprints=False, filter_boards_by_projects=project_id_set
            )

            board_ids = []
            for proj_id in project_id_set:
                board_ids.extend(PROJECT_ID_TO_BOARDS_MAP[proj_id])
            assert len(boards) == len(board_ids)
            for board in boards:
                assert board["id"] in board_ids

            return boards

        _test_filter_by_board({"12"})
        _test_filter_by_board({"12", "51"})
        _test_filter_by_board({"51"})

        all_boards = _test_filter_by_board({"12", "51", "1001"})
        assert len(all_boards) == 3

        no_boards = _test_filter_by_board({"1001"})
        assert len(no_boards) == 0



import dataclasses
import logging
from contextlib import contextmanager
from typing import Optional

import requests

from jf_ingest.events.models import EventState, IngestEvent
from jf_ingest.utils import get_jellyfish_api_base_url, get_jellyfish_api_token

logger = logging.getLogger(__name__)

# NOTE: This code is currently disabled, as we never followed through with the event emission system.


def emit_event_to_jellyfish(event: IngestEvent) -> Optional[requests.Response]:
    """
    Utility function for emitting events directly to Jellyfish, via the Jellyfish app API.

    Args:
        event_name (str): The event name to emit.
        event_state (Optional[EventState], optional): An event state enum to add to the event payload. Defaults to None.
        event_extras (Optional[dict], optional): A dictionary of extra even key/values to join to the event payload. Defaults to None.
        exception (Optional[Exception], optional): An exception object, to mark if this event is related to an exception. Defaults to None.

    Returns:
        Response: This returns what ever the wrapped _func function will return
    """
    if True:
        # TODO: We don't have a good method of getting "global" feature flags in JF Ingest;
        # we currently only have git or jira specific feature flags. We should introduce
        # an endpoint on Jellyfish that can pass forward global feature flags (and also maybe
        # we should redesign git and jira feature flags to do that as well). Feature flags
        # could be cached once they reach JF Ingest, so we don't have to hit the API everytime
        logger.debug(f'Jellyfish Ingest Event Emission is currently disabled')
        return None

    if not get_jellyfish_api_token() or not get_jellyfish_api_base_url():
        logger.error(
            f'emit_event_to_jellyfish was called without jellyfish_api_token or without jellyfish_api_base_url. An event will not be emitted'
        )
        return

    try:
        headers = {"Jellyfish-API-Token": get_jellyfish_api_token()}
        payload = dataclasses.asdict(event)
        response = requests.post(
            f"{get_jellyfish_api_base_url()}/endpoints/ingest/publish-event",
            json=payload,
            headers=headers,
        )
        response.raise_for_status()
        return response
    except Exception as e:
        logger.error(f'Exception {e} encountered when attempting to emit status to Jellyfish')


@contextmanager
def emit_event_to_jellyfish_context_manager(ingest_event: IngestEvent):
    """NOTE: This is deprecated for now. This was originally supposed to be used
    as part of a status emitting system to monitor ingest as it runs. That idea got
    scrapped, however. For now I'm leaving this code in even though it's not referenced anywhere.

    Context Manager for wrapping up logic with a "START" and "END" event emission

    Args:
        event_name (str): The event name to call this logical block of code
        event_extras (Optional[dict], optional): Optional information to submit to the event payload. Defaults to None.
    """
    _exception = None
    try:
        start_event = dataclasses.replace(ingest_event, event_state=EventState.START)
        emit_event_to_jellyfish(start_event)
        yield
    except Exception as e:
        _exception = e
        raise
    finally:
        end_event = dataclasses.replace(
            ingest_event,
            event_state=EventState.ERROR if _exception else EventState.SUCCESS,
            error_message=str(_exception) if _exception else '',
        )
        emit_event_to_jellyfish(end_event)



from contextlib import contextmanager
from unittest.mock import patch

from jf_ingest.config import IngestionConfig
from tests.jf_git.ado.utils import (
    get_adapter,
    get_raw_branches,
    get_raw_change_counts,
    get_raw_commits,
    get_raw_diffs,
    get_raw_iterations,
    get_raw_prs,
    get_raw_repos,
    get_raw_teams,
    get_raw_threads_comments,
    get_raw_users,
)


@contextmanager
def _spoof_all_client_functions():
    raw_prs = get_raw_prs()
    raw_iterations = get_raw_iterations()
    raw_diffs = get_raw_diffs()
    raw_change_counts = get_raw_change_counts()
    raw_commits = get_raw_commits()
    raw_threads = get_raw_threads_comments()
    raw_branches = get_raw_branches()
    raw_repos = get_raw_repos()
    raw_teams = get_raw_teams()
    raw_users = get_raw_users()
    with (
        patch(
            'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pull_requests',
            return_value=raw_prs,
        ),
        patch(
            'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pull_request_iterations',
            return_value=raw_iterations,
        ),
        patch(
            'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pull_request_diff',
            return_value=raw_diffs,
        ),
        patch(
            'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pull_request_changes_counts',
            return_value=raw_change_counts,
        ),
        patch(
            'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pr_commits',
            return_value=raw_commits,
        ),
        patch(
            'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_pr_comment_threads',
            return_value=raw_threads,
        ),
        patch(
            'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_branches',
            return_value=raw_branches,
        ),
        patch(
            'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_all_repos',
            return_value=raw_repos,
        ),
        patch(
            'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_teams',
            return_value=raw_teams,
        ),
        patch(
            'jf_ingest.jf_git.clients.azure_devops.AzureDevopsClient.get_graph_users',
            return_value=raw_users,
        ),
    ):
        yield


def test_ado_smoke_test():
    adapter = get_adapter()
    ingest_config = IngestionConfig(
        jellyfish_api_token='abc',
        company_slug='',
        save_locally=False,
        upload_to_s3=False,
    )
    with _spoof_all_client_functions():
        adapter.load_and_dump_git(adapter.config, ingest_config)



"""Tests for jf_ingest.config module, specifically JiraDownloadConfig feature flag access patterns."""

import pytest

from jf_ingest.config import JiraDownloadConfig
from jf_ingest.constants import Constants


class TestJiraDownloadConfigFeatureFlags:
    """Test feature flag access patterns in JiraDownloadConfig following existing codebase conventions."""

    def test_jql_enhanced_search_enabled_default_false(self):
        """Test that JQL Enhanced Search feature flag defaults to False when not set."""
        config = JiraDownloadConfig(
            company_slug='test',
            url='https://test.atlassian.net',
            gdpr_active=False,
            feature_flags={}
        )
        assert config.feature_flags.get(Constants.JQL_ENHANCED_SEARCH_ENABLED, False) is False

    def test_jql_enhanced_search_enabled_true_when_flag_set(self):
        """Test that JQL Enhanced Search feature flag returns True when enabled."""
        config = JiraDownloadConfig(
            company_slug='test',
            url='https://test.atlassian.net',
            gdpr_active=False,
            feature_flags={Constants.JQL_ENHANCED_SEARCH_ENABLED: True}
        )
        assert config.feature_flags.get(Constants.JQL_ENHANCED_SEARCH_ENABLED, False) is True

    def test_jql_enhanced_search_enabled_false_when_flag_explicitly_disabled(self):
        """Test that JQL Enhanced Search feature flag returns False when explicitly disabled."""
        config = JiraDownloadConfig(
            company_slug='test',
            url='https://test.atlassian.net',
            gdpr_active=False,
            feature_flags={Constants.JQL_ENHANCED_SEARCH_ENABLED: False}
        )
        assert config.feature_flags.get(Constants.JQL_ENHANCED_SEARCH_ENABLED, False) is False

    def test_force_legacy_api_default_false(self):
        """Test that Force Legacy API feature flag defaults to False when not set."""
        config = JiraDownloadConfig(
            company_slug='test',
            url='https://test.atlassian.net',
            gdpr_active=False,
            feature_flags={}
        )
        assert config.feature_flags.get(Constants.FORCE_LEGACY_API, False) is False

    def test_force_legacy_api_true_when_flag_set(self):
        """Test that Force Legacy API feature flag returns True when enabled."""
        config = JiraDownloadConfig(
            company_slug='test',
            url='https://test.atlassian.net',
            gdpr_active=False,
            feature_flags={Constants.FORCE_LEGACY_API: True}
        )
        assert config.feature_flags.get(Constants.FORCE_LEGACY_API, False) is True

    def test_force_legacy_api_false_when_flag_explicitly_disabled(self):
        """Test that Force Legacy API feature flag returns False when explicitly disabled."""
        config = JiraDownloadConfig(
            company_slug='test',
            url='https://test.atlassian.net',
            gdpr_active=False,
            feature_flags={Constants.FORCE_LEGACY_API: False}
        )
        assert config.feature_flags.get(Constants.FORCE_LEGACY_API, False) is False

    def test_both_feature_flags_work_together(self):
        """Test that both feature flags can be set independently using the standard pattern."""
        config = JiraDownloadConfig(
            company_slug='test',
            url='https://test.atlassian.net',
            gdpr_active=False,
            feature_flags={
                Constants.JQL_ENHANCED_SEARCH_ENABLED: True,
                Constants.FORCE_LEGACY_API: False
            }
        )
        assert config.feature_flags.get(Constants.JQL_ENHANCED_SEARCH_ENABLED, False) is True
        assert config.feature_flags.get(Constants.FORCE_LEGACY_API, False) is False

    def test_feature_flags_with_none_values(self):
        """Test that feature flags handle None values gracefully using the standard pattern."""
        config = JiraDownloadConfig(
            company_slug='test',
            url='https://test.atlassian.net',
            gdpr_active=False,
            feature_flags={
                Constants.JQL_ENHANCED_SEARCH_ENABLED: None,
                Constants.FORCE_LEGACY_API: None
            }
        )
        # None values should return None, not the default, when key exists
        assert config.feature_flags.get(Constants.JQL_ENHANCED_SEARCH_ENABLED, False) is None
        assert config.feature_flags.get(Constants.FORCE_LEGACY_API, False) is None

    def test_feature_flags_with_no_feature_flags_dict(self):
        """Test that feature flag access works when feature_flags uses default empty dict."""
        config = JiraDownloadConfig(
            company_slug='test',
            url='https://test.atlassian.net',
            gdpr_active=False
        )
        assert config.feature_flags.get(Constants.JQL_ENHANCED_SEARCH_ENABLED, False) is False
        assert config.feature_flags.get(Constants.FORCE_LEGACY_API, False) is False

    def test_feature_flag_constants_match_expected_values(self):
        """Test that the constants have the expected values as specified in the task."""
        assert Constants.JQL_ENHANCED_SEARCH_ENABLED == 'makara-jql-enhanced-search-enabled-2025Q3'
        assert Constants.FORCE_LEGACY_API == 'makara-force-legacy-api-2025Q3'

    def test_feature_flags_follow_existing_pattern(self):
        """Test that the new feature flags follow the same pattern as existing ones."""
        config = JiraDownloadConfig(
            company_slug='test',
            url='https://test.atlassian.net',
            gdpr_active=False,
            feature_flags={
                Constants.JQL_ENHANCED_SEARCH_ENABLED: True,
                Constants.FORCE_LEGACY_API: False,
                Constants.PULL_SPRINTS_BY_ID: True,  # Existing feature flag for comparison
            }
        )
        
        # Test that all feature flags use the same access pattern
        assert config.feature_flags.get(Constants.JQL_ENHANCED_SEARCH_ENABLED, False) is True
        assert config.feature_flags.get(Constants.FORCE_LEGACY_API, False) is False
        assert config.feature_flags.get(Constants.PULL_SPRINTS_BY_ID, False) is True


import logging
import os
import sys
from datetime import datetime

import pytz
import urllib3
import yaml

from jf_ingest.config import (
    AzureDevopsAuthConfig,
    GitAuthConfig,
    GitConfig,
    GitLabAuthConfig,
    GitProvider,
)
from jf_ingest.constants import Constants
from jf_ingest.jf_git.adapters import load_and_push_git_to_s3
from jf_ingest.jf_jira import IngestionConfig, load_and_push_jira_to_s3
from jf_ingest.jf_jira.auth import JiraDownloadConfig, get_jira_connection
from jf_ingest.jf_jira.custom_fields import identify_custom_field_mismatches
from jf_ingest.jf_jira.downloaders import is_jql_enhanced_search_available
from jf_ingest.validation import validate_git, validate_jira

logger = logging.getLogger(__name__)

# Start of Epoch Time
_default_datetime_str = '1970-01-01'


def setup_harness_logging(logging_level: int):
    """Helper function to setting up logging in the harness"""
    logging.basicConfig(
        level=logging_level,
        format=(
            "%(asctime)s %(threadName)s %(levelname)s %(name)s %(message)s"
            if logging_level == logging.DEBUG
            else "%(message)s"
        ),
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logging.getLogger(urllib3.__name__).setLevel(logging.WARNING)


def _process_jira_config(general_config_data: dict) -> JiraDownloadConfig:
    jira_config_data = general_config_data["jira_config"]
    # Translate Datetimes first
    jira_config_data["work_logs_pull_from"] = datetime.strptime(
        jira_config_data.get("work_logs_pull_from", _default_datetime_str), "%Y-%m-%d"
    )
    jira_config_data["pull_from"] = datetime.strptime(
        jira_config_data.get("pull_from", _default_datetime_str), "%Y-%m-%d"
    )
    jira_config_data["project_id_to_pull_from"] = {
        project_id: datetime.strptime(pull_from, "%Y-%m-%d")
        for project_id, pull_from in jira_config_data.get("project_id_to_pull_from", {}).items()
    }
    if not (jira_url := os.getenv("JIRA_URL")):
        raise Exception(
            f'JIRA_URL not detected as an environment variable! Is it set in your creds.env?'
        )
    jira_config_data['url'] = jira_url
    # Generate object in memory
    jira_config = JiraDownloadConfig(**jira_config_data)
    if not (jira_username := os.getenv("JIRA_USERNAME")):
        raise Exception(
            f'JIRA_USERNAME not detected as an environment variable! Is it set in your creds.env?'
        )
    if not (jira_password := os.getenv("JIRA_PASSWORD")):
        raise Exception(
            f'JIRA_PASSWORD not detected as an environment variable! Is it set in your creds.env?'
        )
    jira_config.user = jira_username
    jira_config.password = jira_password
    return jira_config


def _process_git_configs(general_config_data: dict) -> list[GitConfig]:
    git_config_data = general_config_data["git_configs"][0]
    company_slug = git_config_data['company_slug']
    # Process 'default' time for repos and commits
    git_config_data["pull_from"] = datetime.strptime(
        git_config_data.get("pull_from", _default_datetime_str),
        "%Y-%m-%d",
    ).replace(tzinfo=pytz.UTC)

    # Translate PRs
    for repo_id, date_str in git_config_data[f'repos_to_prs_last_updated'].items():
        git_config_data[f'repos_to_prs_last_updated'][repo_id] = datetime.strptime(
            date_str, "%Y-%m-%d"
        ).replace(tzinfo=pytz.UTC)

    # Create separate auth config
    git_provider = GitProvider(git_config_data['git_provider'].upper())
    if git_provider == GitProvider.GITHUB:
        auth_config = GitAuthConfig(
            company_slug=company_slug,
            token=os.getenv('GITHUB_TOKEN'),
            base_url='',
            verify=False,
            session=None,
        )
    elif git_provider == GitProvider.ADO:
        auth_config = AzureDevopsAuthConfig(
            company_slug=company_slug,
            token=os.getenv(f'GITHUB_TOKEN'),
            base_url='',
            verify=True,
            session=None,
        )
    elif git_provider == GitProvider.GITLAB:
        auth_config = GitLabAuthConfig(
            company_slug=company_slug,
            token=os.getenv(f'GITHUB_TOKEN'),
            base_url='',
            verify=True,
            session=None,
        )
    else:
        raise Exception(f'{git_provider} not recognized in harness init')
    return [GitConfig(git_auth_config=auth_config, **git_config_data)]


if __name__ == "__main__":
    """
    NOTE: This is a work in progress developer debugging tool.
    it is currently run by using the following command:
       pdm run ingest_harness [--debug]
    and it requires you to have a creds.env and a config.yml file at
    the root of this project
    """
    debug_mode = "--debug" in sys.argv
    validate_mode = "--validate" in sys.argv
    run_git_harness = '--git' in sys.argv
    run_jira_harness = '--jira' in sys.argv
    skip_custom_field_detect_and_repair = '--skip-field-repair' in sys.argv

    if not run_git_harness and not run_jira_harness:
        # If neither the git or jira arg are supplied, run validation for both
        run_git_harness = True
        run_jira_harness = True

    # Get Config data for Ingestion Config
    with open("./config.yml") as yaml_file:
        yaml_data = yaml.safe_load(yaml_file)
        general_config_data = yaml_data["general"]
        if not (company_slug := os.getenv('COMPANY_SLUG')):
            raise Exception(f'COMPANY_SLUG not found as an env variable. Is it set in creds.env?')
        if not (api_token := os.getenv("JELLYFISH_API_TOKEN")):
            raise Exception(f'COMPANY_SLUG not found as an env variable. Is it set in creds.env?')
        ingest_config = IngestionConfig(
            timestamp=datetime.utcnow().strftime("%Y%m%d_%H%M%S"),
            jellyfish_api_token=api_token,
            **general_config_data,
            company_slug=company_slug,
        )
        ingest_config.local_file_path = f"{ingest_config.local_file_path}/{ingest_config.timestamp}"
        print(f'Saving data locally to {ingest_config.local_file_path}')

        # Processes Jira Info
        if run_jira_harness:
            if 'jira_config' in general_config_data:
                ingest_config.jira_config = _process_jira_config(general_config_data)
            else:
                logger.warning(
                    f'Attempted to run Jira Validation but the Jira Config was not found in the config.yml file'
                )

        # Process Git Info
        if run_git_harness:
            if 'git_configs' in general_config_data:
                ingest_config.git_configs = _process_git_configs(general_config_data)
            else:
                logger.warning(
                    f'Attempting to run Git Validation but the Git Config was not found in the config.yml file'
                )

    setup_harness_logging(logging_level=logging.DEBUG if debug_mode else logging.INFO)

    if validate_mode:
        if run_jira_harness and ingest_config.jira_config:
            validate_jira(ingest_config.jira_config)
        if run_git_harness and 'git_configs' in general_config_data:
            validate_git(ingest_config.git_configs)
    else:
        if run_jira_harness and ingest_config.jira_config:
            # NOTE: The config file stores this value as a list, when it should actually be a set.
            # This is a quick and cheap fix to re-cast this
            ingest_config.jira_config.jellyfish_issue_ids_for_redownload = set(
                ingest_config.jira_config.jellyfish_issue_ids_for_redownload
            )
            if not skip_custom_field_detect_and_repair:
                # Test if JQL Enhanced Search is available
                use_jql_enhanced_search = is_jql_enhanced_search_available(
                    jira_config=ingest_config.jira_config,
                    jql_enhanced_search_enabled=ingest_config.jira_config.feature_flags.get(
                        Constants.JQL_ENHANCED_SEARCH_ENABLED, False
                    ),
                    force_legacy_api=ingest_config.jira_config.feature_flags.get(
                        Constants.FORCE_LEGACY_API, False
                    ),
                )

                jcfv_update_payload = identify_custom_field_mismatches(
                    ingest_config=ingest_config, use_jql_enhanced_search=use_jql_enhanced_search
                )
                logger.info("Submitting list of issue IDs to repair to Jellyfish")

                # TODO: All this stuff should be moved INTO the identify_custom_field_mismatches function.
                # The identify_custom_field_mismatches should post the list of all_ids_to_redownload to Jellyfish,
                # and then Jira Download should pull that list
                missing_db_ids = [
                    jcfv.jira_issue_id for jcfv in jcfv_update_payload.missing_from_db_jcfv
                ]
                missing_jira_ids = [
                    jcfv.jira_issue_id for jcfv in jcfv_update_payload.missing_from_jira_jcfv
                ]
                missing_out_of_sync_ids = [
                    jcfv.jira_issue_id for jcfv in jcfv_update_payload.out_of_sync_jcfv
                ]

                all_ids = set(missing_db_ids + missing_jira_ids + missing_out_of_sync_ids)

                all_ids_to_redownload = set([str(id) for id in all_ids])

                ingest_config.jira_config.jellyfish_issue_ids_for_redownload.update(
                    all_ids_to_redownload
                )
                logger.info(f'----')
                logger.info(
                    f'Done detecting customfield mismatches, {len(all_ids_to_redownload)} additional IDs will be redownloaded in Jira Download.'
                )
            logger.info('Beginning Jira Download')
            logger.info(f'----')

            jira_success = load_and_push_jira_to_s3(ingest_config)
            if not jira_success:
                raise RuntimeError(
                    f'Jira Download function {load_and_push_jira_to_s3.__name__} did not return True, which indicates a failure!'
                )
        if run_git_harness:
            load_and_push_git_to_s3(ingest_config)



import json
import os
import requests_mock
from typing import Optional
from unittest import TestCase
from unittest.mock import MagicMock

from jf_ingest.config import IngestionConfig, IngestionType, JiraDownloadConfig
from jf_ingest.jf_jira import custom_fields as custom_fields
from jf_ingest.jf_jira.custom_fields import (
    JELLYFISH_CUSTOM_FIELDS_ENDPOINT,
    JELLYFISH_MARK_FOR_REDOWNLOAD_ENDPOINT,
    JCFVUpdate
)
from tests.jf_jira.utils import get_fixture_file_data

ATLASSIAN_BASE_URL = 'mock://test-co.atlassian.net/'
JELLYFISH_BASE_URL = 'mock://jellyfish.jellyfish'


class TestCustomFieldsComparison(TestCase):
    def setUp(self):
        self.missing_from_jira_jcfv = JCFVUpdate(
            field_id='201',
            field_key='customfield_301',
            field_type='team',
            jira_issue_id='101',
            value_new=None,
            value_old={"name": "T2"}
        )
        self.missing_from_db_jcfv = JCFVUpdate(
            field_id='201',
            field_key='customfield_301',
            field_type='team',
            jira_issue_id='100',
            value_new={
                'self': 'https://test-co.atlassian.net/rest/api/2/customFieldOption/12347',
                'value': 'NEW TEAM 10',
                'id': '12347',
            },
            value_old=None
        )
        self.out_of_sync_jcfv = JCFVUpdate(
            field_id='200',
            field_key='customfield_300',
            field_type='team',
            jira_issue_id='100',
            value_new={
                'self': 'https://test-co.atlassian.net/rest/api/2/customFieldOption/12345',
                'value': 'NEW TEAM 1',
                'id': '12345',
            },
            value_old={"name": "T1"}
        )

    def _get_ingestion_config(self):
        company_slug = "test_company"
        url = ATLASSIAN_BASE_URL
        return IngestionConfig(
            company_slug=company_slug,
            jira_config=JiraDownloadConfig(
                company_slug=company_slug,
                url=url,
                personal_access_token='pat',
                gdpr_active=True,
            ),
            git_configs=[],
            jellyfish_api_token='some_token',
            jellyfish_api_base=JELLYFISH_BASE_URL,
            ingest_type=IngestionType.DIRECT_CONNECT,
        )


    def _get_jellyfish_custom_fields_default_response(self) -> dict:
        return json.loads(get_fixture_file_data(fixture_path='jellyfish_custom_fields.json'))


    def _get_jellyfish_custom_fields_empty_response(self) -> dict:
        return json.loads(
            get_fixture_file_data(fixture_path='jellyfish_custom_fields_empty_response.json')
        )


    def _mock_synchronous_thread(self):
        # Mock out ThreadPoolExecutor to get a synchronous run
        def _synchronous_run(f, *args, **kwargs):
            f(*args, **kwargs)
            thread_obj = MagicMock()
            thread_obj.running = MagicMock()
            thread_obj.running.return_value = False
            return thread_obj

        synchronous_run_mock = MagicMock()
        custom_fields.ThreadPoolExecutor = MagicMock()
        custom_fields.ThreadPoolExecutor.return_value = synchronous_run_mock
        synchronous_run_mock.submit = _synchronous_run

        return synchronous_run_mock


    def _mock_custom_field_responses(self, issues_fixture_path: Optional[str] = 'issues_for_custom_fields.json'):
        # Mock out Jira response
        issues = get_fixture_file_data(
            fixture_path=os.path.join('api_responses', issues_fixture_path)
        )
        issues_resp = json.loads(issues)['issues']
        custom_fields.pull_jira_issues_by_jira_ids = MagicMock()
        custom_fields.pull_jira_issues_by_jira_ids.return_value = issues_resp

        # Mock out Jira connection and batch size request
        custom_fields.get_jira_connection = MagicMock()
        custom_fields.get_jira_search_batch_size = MagicMock()
        custom_fields.get_jira_search_batch_size.return_value = 100


    def test_custom_fields_comparison_base_case(self):
        config = self._get_ingestion_config()
        jf_response = self._get_jellyfish_custom_fields_default_response()

        with requests_mock.Mocker() as mocker:
            mocker.register_uri(
                'GET',
                f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}?cursor=0&limit={custom_fields.JELLYFISH_CUSTOM_FIELDS_API_LIMIT}",
                json=jf_response,
                status_code=200,
            )

            mocker.register_uri(
                'GET',
                f"{config.jellyfish_api_base}/endpoints/jira/issues/count",
                json={
                    'total_issues_in_jellyfish': 0
                },  # This is just for TQDM UX, and not really important for testing
                status_code=200,
            )

            redownload_mock = mocker.register_uri(
                'POST',
                f"{config.jellyfish_api_base}/{JELLYFISH_MARK_FOR_REDOWNLOAD_ENDPOINT}",
                json={},
                status_code=200,
            )

            self._mock_synchronous_thread()
            self._mock_custom_field_responses()

            # Generate update payload
            update_payload = custom_fields.identify_custom_field_mismatches(config, nthreads=1)

            self.assertEqual(len(update_payload.missing_from_jira_jcfv), 1)
            self.assertEqual(len(update_payload.missing_from_db_jcfv), 1)
            self.assertEqual(len(update_payload.out_of_sync_jcfv), 1)

            # Missing from Jira (DELETE)
            self.assertEqual(update_payload.missing_from_jira_jcfv[0],self.missing_from_jira_jcfv)

            # Missing from DB (INSERT)
            self.assertEqual(update_payload.missing_from_db_jcfv[0], self.missing_from_db_jcfv)

            # Out-of-sync (UPDATE)
            self.assertEqual(update_payload.out_of_sync_jcfv[0], self.out_of_sync_jcfv)

            self.assertFalse(redownload_mock.called)


    def test_custom_fields_mark_for_redownload(self):
        config = self._get_ingestion_config()
        jf_response = self._get_jellyfish_custom_fields_default_response()

        with requests_mock.Mocker() as mocker:
            mocker.register_uri(
                'GET',
                f"{config.jellyfish_api_base}/{JELLYFISH_CUSTOM_FIELDS_ENDPOINT}?cursor=0&limit={custom_fields.JELLYFISH_CUSTOM_FIELDS_API_LIMIT}",
                json=jf_response,
                status_code=200,
            )

            mocker.register_uri(
                'GET',
                f"{config.jellyfish_api_base}/endpoints/jira/issues/count",
                json={
                    'total_issues_in_jellyfish': 0
                },  # This is just for TQDM UX, and not really important for testing
                status_code=200,
            )

            redownload_mock = mocker.register_uri(
                'POST',
                f"{config.jellyfish_api_base}/{JELLYFISH_MARK_FOR_REDOWNLOAD_ENDPOINT}",
                [
                    {'json': {}, 'status_code': 504},
                    {'json': {}, 'status_code': 200},
                ]
            )

            self._mock_synchronous_thread()
            self._mock_custom_field_responses()

            # Generate update payload
            update_payload = custom_fields.identify_custom_field_mismatches(config, nthreads=1, mark_for_redownload=True)

            self.assertEqual(len(update_payload.missing_from_jira_jcfv), 1)
            self.assertEqual(len(update_payload.missing_from_db_jcfv), 1)
            self.assertEqual(len(update_payload.out_of_sync_jcfv), 1)
            self.assertEqual(update_payload.missing_from_jira_jcfv[0], self.missing_from_jira_jcfv)
            self.assertEqual(update_payload.missing_from_db_jcfv[0], self.missing_from_db_jcfv)
            self.assertEqual(update_payload.out_of_sync_jcfv[0], self.out_of_sync_jcfv)

            # We should attempt twice with the same payload. A 504 triggers retry. Issue id order does not matter.
            self.assertEqual(redownload_mock.call_count, 2)
            self.assertCountEqual(redownload_mock.request_history[0].json()['issue_ids'], ['100', '101'])
            self.assertCountEqual(redownload_mock.request_history[1].json()['issue_ids'], ['101', '100'])



import requests
from jf_ingest.constants import Constants


PATH_TO_TEST_FIXTURES = 'tests/jf_git/gitlab/fixtures'

TEST_COMPANY_SLUG = 'A-Company'
TEST_BASE_URL = 'https://www.a-website.com'
TEST_BASE_GQL_URL = f'{TEST_BASE_URL}/api/graphql'
TEST_TOKEN = 'A Spoofed Token'
TEST_ORG_LOGIN = '1'
TEST_FULL_PATH = 'test-full-path'
TEST_INSTANCE_SLUG = 'a-test-instance-slug'
TEST_INSTANCE_FILE_KEY = 'a-test-file-key'
EXPECTED_AUTH_HEADER = {
    'Authorization': f'Bearer {TEST_TOKEN}',
    'Content-Type': 'application/json',
    'User-Agent': f'{Constants.JELLYFISH_USER_AGENT} ({requests.utils.default_user_agent()})',
}


from contextlib import contextmanager
import json

import requests_mock

from jf_ingest.jf_jira.downloaders import download_global_components
from tests.jf_jira.utils import (
    _register_jira_uri,
    get_fixture_file_data,
    get_jira_mock_connection,
)


@contextmanager
def _mock_components_endpoint(
    m: requests_mock.Mocker,
    endpoint_base: str,
):
    component_1 = get_fixture_file_data(f"api_responses/components_1.json")
    component_2 = get_fixture_file_data(f"api_responses/components_2.json")
    print(endpoint_base)
    _register_jira_uri(
        m,
        endpoint=f"component?startAt=0&maxResults=50",
        return_value=component_1,
    )
    _register_jira_uri(
        m,
        endpoint=f"component?startAt=50&maxResults=50",
        return_value=component_2,
    )
    yield

def test_download_board_with_project_filtering_smoke_test():
    base_url = "https://test-co.atlassian.net"
    with requests_mock.Mocker() as m:
        jira_conn = get_jira_mock_connection(mocker=m)
        with _mock_components_endpoint(m, base_url):
            components = download_global_components(
                jira_connection=jira_conn,
            )
            
            component_fixture_data = json.loads(get_fixture_file_data(f"api_responses/components_1.json"))['values'] + json.loads(get_fixture_file_data(f"api_responses/components_2.json"))['values']
            assert len(components) == len(
                [c for c in component_fixture_data if 'ari' in c]
            )



import json
from contextlib import contextmanager
from datetime import datetime

import requests_mock

from jf_ingest.jf_jira.downloaders import (
    _convert_datetime_to_worklog_timestamp,
    download_worklogs,
)
from tests.jf_jira.utils import (
    _register_jira_uri,
    get_fixture_file_data,
    get_jira_mock_connection,
)

DELETED_WORKLOG_IDS = [10007, 10008]


@contextmanager
def worklogs_context_manager(existing_ids: list[int] = []):
    with requests_mock.Mocker() as m:
        # Get specific updated since data
        worklog_updated_data_str = get_fixture_file_data("api_responses/worklog_updated.json")
        _register_jira_uri(
            m, endpoint="worklog/updated?since=0", return_value=worklog_updated_data_str
        )
        # List Updated Since Data
        worklog_list_data_str = get_fixture_file_data("api_responses/worklog_list.json")
        worklog_filtered = [
            wl for wl in json.loads(worklog_list_data_str) if int(wl["issueId"]) in existing_ids
        ]
        _register_jira_uri(
            m,
            endpoint="worklog/list",
            return_value=json.dumps(worklog_filtered),
            HTTP_ACTION="POST",
        )
        # Get deleted
        worklog_deleted_data_str = get_fixture_file_data("api_responses/worklog_deleted.json")
        _register_jira_uri(
            m, endpoint="worklog/deleted?since=0", return_value=worklog_deleted_data_str
        )
        yield


def test_download_worklogs_schema():
    """
    worklogs is a simple function, so we just need a simple test
    """
    with worklogs_context_manager():
        worklog_dict = download_worklogs(get_jira_mock_connection(), [], datetime.min)

        # Assert schema of worklog dict object
        assert type(worklog_dict) == dict
        # Check existing
        assert "existing" in worklog_dict
        assert type(worklog_dict["existing"]) == list
        # Check deleted
        assert "deleted" in worklog_dict
        assert type(worklog_dict["deleted"]) == list


def test_download_worklogs_no_issues_to_search():
    """
    worklogs is a simple function, so we just need a simple test
    """
    with worklogs_context_manager():
        worklog_dict = download_worklogs(get_jira_mock_connection(), [], datetime.min)

        assert len(worklog_dict["existing"]) == 0
        assert worklog_dict["deleted"] == DELETED_WORKLOG_IDS


def test_download_worklogs_no_issues_to_search():
    """
    worklogs is a simple function, so we just need a simple test
    """
    ids = [11143, 11201]
    with worklogs_context_manager(ids):
        worklog_dict = download_worklogs(get_jira_mock_connection(), ids, datetime.min)

        assert len(worklog_dict["existing"]) == 4
        assert worklog_dict["deleted"] == DELETED_WORKLOG_IDS


def test_convert_datetime_to_worklog_timestamp():
    since_timestamp = _convert_datetime_to_worklog_timestamp(datetime.min)
    assert since_timestamp == 0

    a_day = datetime(1995, 12, 29)
    a_day_since_timestamp = _convert_datetime_to_worklog_timestamp(a_day)
    assert a_day.timestamp() * 1000 == a_day_since_timestamp

    b_day = datetime(1995, 12, 29, 3)
    b_day_since_timestamp = _convert_datetime_to_worklog_timestamp(b_day)
    assert b_day.timestamp() * 1000 == b_day_since_timestamp



import logging
from typing import Optional, Union

import requests
from jira import JIRA, JIRAError
from jira.resources import GreenHopperResource
from requests.auth import AuthBase

from jf_ingest.config import JiraAuthConfig, JiraAuthMethod, JiraDownloadConfig
from jf_ingest.utils import retry_for_status

logger = logging.getLogger(__name__)


class JiraAuthenticationException(Exception):
    def __init__(self, *args, original_exception: Optional[Exception] = None):
        self.original_exception = original_exception
        super().__init__(*args)


# HACK(asm,2024-05-10): We are seeing questionable HTTP 401 responses from Jira that seem to be
# related to cookie-based authentication validation. Specifically, we believe that requests made
# using a bearer token are creating an expiring session on the server side, and it is causing Jira
# to (incorrectly) challenge us to re-authenticate based on this expired session. To avoid this, we
# are customizing the authentication method so that we never send cookies when we are using bearer
# token authentication. If this works, this should be upstreamed.
class TokenAuth(AuthBase):
    def __init__(self, token: str):
        self._token = token

    def __call__(self, r: requests.PreparedRequest) -> requests.PreparedRequest:
        # never send cookies for token auth
        r.headers["Cookie"] = ''
        r.headers["authorization"] = f"Bearer {self._token}"
        return r


def get_jira_connection(
    config: Union[JiraAuthConfig, JiraDownloadConfig],
    auth_method: JiraAuthMethod = JiraAuthMethod.BasicAuth,
    max_retries=3,
    use_jql_enhanced_search: bool = False,
) -> JIRA:
    """Get a JIRA connection object

    Args:
        config (JiraAuthConfig | JiraDownloadConfig): A JIRA Configuration Object that contains authentication information
        auth_method (JiraAuthMethod, optional): The Authentication method used. BasicAuth and AtlassianConnect are currently supported. Defaults to JiraAuthMethod.BasicAuth.
        max_retries (int, optional): The retry limit used by the JiraConnection.ResilientSession object. Defaults to 3.
        use_jql_enhanced_search (bool, optional): Whether to configure connection for JQL Enhanced Search API. When True, uses REST API v3 for /search/jql endpoints. When False, uses REST API v2 for backward compatibility. Defaults to False.

    Raises:
        Several Errors can be raised by this class. Please ensure your config is properly set up

    Returns:
        JIRA: A JIRA connection object
    """
    should_ssl_verify = not config.bypass_ssl_verification

    # This is the base Jira Connection KWARGs. For different
    # authentication methods, we will need to add specific fields
    jira_conn_kwargs = {
        "timeout": 600,
        "server": config.url,
        "max_retries": max_retries,
        "options": {
            "agile_rest_path": GreenHopperResource.AGILE_BASE_REST_PATH,
            "verify": should_ssl_verify,
            "headers": {
                "Cache-Control": "no-cache",
                "Content-Type": "application/json",
                "Accept": "application/json,*/*;q=0.9",
                "X-Atlassian-Token": "no-check",
            },
        },
    }

    # Determine REST API version based on intended usage
    if use_jql_enhanced_search:
        jira_conn_kwargs['options']['rest_api_version'] = '3'

    using_token_auth = False

    if auth_method.value == JiraAuthMethod.AtlassianConnect.value:
        if not config.connect_app_active:
            raise RuntimeError(
                f'Atlassian connect integration for {config.company_slug} is disabled. Check "connect_app_active" flag state.'
            )

        # The customer has installed our Atlassian Connect app; we authenticate to the Jira
        # API using the JWT attributes
        shared_secret = config.jwt_attributes.get("sharedSecret")
        client_key = config.jwt_attributes.get("clientKey")
        key = config.jwt_attributes.get("key")

        if not (shared_secret and client_key and key):
            raise RuntimeError(
                f"Atlassian connect integration for {config.company_slug} is misconfigured"
            )

        logger.info(f"Authenticating to Jira API at {config.url} using JWT attributes")
        jira_conn_kwargs['jwt'] = {
            "secret": shared_secret,
            "payload": {"iss": key, "aud": client_key},
        }
        jira_conn: JIRA = retry_for_status(
            JIRA, max_retries_for_retry_for_status=3, **jira_conn_kwargs
        )

    elif auth_method.value == JiraAuthMethod.BasicAuth.value:
        # Attempt to log a helpful warning if a customer has a Jira User provided but NOT a jira password provided
        if config.user and not config.password:
            optional_bearer_token_message = (
                f'A Jira Bearer token was being detected as being provided, so we will attempt to authenticate with that instead. '
                'If you would like to authenticate with only the Bear token please remove the Jira User field provided, as it is not necessary.'
            )
            logger.warning(
                f'Jira Authentication has detected that a Jira User value has been provided, '
                'but that Jira Password value was left blank. '
                'If you would like to authenticate using a User/Password scheme, please provide the username and password. '
                f'{optional_bearer_token_message if config.personal_access_token else ""}'
            )

        # Attempt user/password auth only if we have both a user and a password
        if config.user and config.password:
            logger.info(
                f"Authenticating to Jira API at {config.url} "
                f"using the username and password secrets for {config.user} of company {config.company_slug}"
            )
            # Add in basic auth info
            jira_conn_kwargs['basic_auth'] = (config.user, config.password)
        elif config.personal_access_token:
            logger.info(
                f"Authenticating to Jira API at {config.url} "
                f"using the personal_access_token secret for {config.company_slug}"
            )
            # Add in personal access token
            jira_conn_kwargs['token_auth'] = config.personal_access_token

            using_token_auth = True
        else:
            raise RuntimeError(
                f"No valid basic authentication mechanism for {config.url} - need a username/password combo or a personal access token"
            )

        while True:
            try:
                jira_conn = retry_for_status(
                    JIRA, max_retries_for_retry_for_status=3, **jira_conn_kwargs
                )
                break
            except JIRAError as e:  # catch generic error raised from JIRA
                if hasattr(e, "status_code") and e.status_code in (401, 403):
                    raise JiraAuthenticationException(
                        f"Jira authentication (HTTP ERROR CODE {e.status_code}) failed for {config.url} using {config.user} and the jira_password secret for {config.company_slug}",
                        original_exception=e,
                    )
                else:
                    raise
    else:
        raise RuntimeError(
            f"No valid authentication mechanism for {config.url} (Auth Method: {auth_method})."
        )

    # set user-agent
    jira_conn._session.headers["User-Agent"] = (
        f'{config.user_agent} ({jira_conn._session.headers["User-Agent"]})'
    )

    # HACK(asm,2024-05-10): See comment above `TokenAuth` class definition - overriding
    # authentication here to force no cookies.
    if using_token_auth and config.personal_access_token:
        jira_conn._session.auth = TokenAuth(config.personal_access_token)

    return jira_conn



import json
from unittest.mock import MagicMock, patch

import pytest
import requests
import requests_mock
from requests_mock.exceptions import NoMockAddress

from jf_ingest.config import GitLabAuthConfig
from jf_ingest.constants import Constants
from jf_ingest.jf_git.clients.gitlab import GitlabClient
from jf_ingest.utils import RetryLimitExceeded
from tests.jf_git.gitlab.utils import (
    EXPECTED_AUTH_HEADER,
    TEST_BASE_GQL_URL,
    TEST_BASE_URL,
    TEST_COMPANY_SLUG,
    TEST_TOKEN,
    get_fixture_data,
    spoof_organizations_through_gql,
    spoof_repositories_through_gql,
)


def test_gitlab_client_constructor():
    auth_config = GitLabAuthConfig(
        company_slug=TEST_COMPANY_SLUG,
        base_url=TEST_BASE_URL,
        token=TEST_TOKEN,
        verify=False,
    )
    with patch.object(GitlabClient, 'get_api_version', return_value='17.4'):
        client = GitlabClient(auth_config)

    assert client.company_slug == auth_config.company_slug
    assert client.gql_base_url == f'{auth_config.base_url}/api/graphql'
    assert client.rest_api_url == auth_config.base_url
    assert client.session.headers['Authorization'] == f'Bearer {auth_config.token}'
    assert client.session.headers['Content-Type'] == 'application/json'


def test_gitlab_client_compat_configuration():
    auth_config = GitLabAuthConfig(
        company_slug=TEST_COMPANY_SLUG,
        base_url=TEST_BASE_URL,
        token=TEST_TOKEN,
        verify=False,
    )
    # This is list of tuples containing the version to test and a value for gql_skip_pr_closed_at.
    # We may want to spoof full configurations as more keys are added.
    test_values = [('17.4.0', False), ('17.4.0-pre', False), ('17.6.1', False), ('16.4.0-ee', True), ('16.4.0', True), ('garbage_input', True)]
    for test_value in test_values:
        with patch.object(GitlabClient, 'get_api_version', return_value=test_value[0]):
            client = GitlabClient(auth_config)
        assert client.compatibility_config['gql_skip_pr_closed_at'] == test_value[1]

def test_gitlab_client_constructor_with_passed_in_session():
    spoofed_session = requests.Session()
    spoofed_header_key = 'Spoofed-Header'
    spoofed_header_value = 123
    spoofed_session.headers.update({
        spoofed_header_key: spoofed_header_value
    })
    auth_config = GitLabAuthConfig(
        company_slug='A-Company',
        base_url='www.company.net',
        token='A Spoofed Token',
        verify=False,
        session=spoofed_session
    )
    with patch.object(GitlabClient, 'get_api_version', return_value='17.4'):
        client = GitlabClient(auth_config)

    assert client.company_slug == auth_config.company_slug
    assert client.gql_base_url == f'{auth_config.base_url}/api/graphql'
    assert client.rest_api_url == auth_config.base_url
    assert client.session.headers[spoofed_header_key] == spoofed_header_value



@pytest.fixture()
def client():
    auth_config = GitLabAuthConfig(
        company_slug=TEST_COMPANY_SLUG,
        base_url=TEST_BASE_URL,
        token=TEST_TOKEN,
        verify=False,
    )
    with patch.object(GitlabClient, 'get_api_version', return_value='17.4'):
        return GitlabClient(auth_config)

def test_get_raw_gql_result_simple(client: GitlabClient, requests_mock: requests_mock.Mocker):
    # Test Data
    test_query_body = "test_query_body"
    # Mock classes/data
    session_mock_caller_data = {'query': test_query_body}
    session_mock_return_data = {'data': {'test_key': 'test_value'}}
    requests_mock.post(url=TEST_BASE_GQL_URL, json=session_mock_return_data, request_headers=EXPECTED_AUTH_HEADER)

    # call function
    returned_data = client.get_raw_result_gql(query_body=test_query_body)

    assert returned_data == session_mock_return_data
    assert requests_mock.last_request.json() == session_mock_caller_data

def test_get_organization_full_path(client: GitlabClient, requests_mock: requests_mock.Mocker):
    raw_data = get_fixture_data('raw_organization_from_rest_api.json')
    group_id = raw_data['id']
    requests_mock.get(
        url=f'{TEST_BASE_URL}/api/v4/groups/{group_id}?with_projects=False', 
        json=raw_data, 
        request_headers=EXPECTED_AUTH_HEADER
    )
    
    name, full_path_response, web_url = client.get_organization_name_full_path_and_url(login=group_id)
    
    assert name == raw_data['name']
    assert full_path_response == raw_data['full_path']
    assert web_url == raw_data['web_url']
    
def test_get_raw_gql_results_retries_on_429s(mocker, client: GitlabClient, requests_mock: requests_mock.Mocker):
    attempts = 10
    requests_mock.post(url=TEST_BASE_GQL_URL, request_headers=EXPECTED_AUTH_HEADER, status_code=429)
    with mocker.patch('time.sleep', return_value=None), pytest.raises(RetryLimitExceeded):
        client.get_raw_result_gql('', max_attempts=attempts)
        assert requests_mock.call_count == attempts

def test_get_raw_gql_result_verify_headers_present(client: GitlabClient, requests_mock: requests_mock.Mocker):
    # Test Data
    test_query_body = "test_query_body"
    # Mock classes/data
    session_mock_caller_data = {'query': test_query_body}
    session_mock_return_data = {'data': {'test_key': 'test_value'}}
    requests_mock.post(url=TEST_BASE_GQL_URL, json=session_mock_return_data, request_headers=EXPECTED_AUTH_HEADER)

    # call function
    returned_data = client.get_raw_result_gql(query_body=test_query_body)

    assert returned_data == session_mock_return_data
    assert requests_mock.last_request.json() == session_mock_caller_data
 
def test_get_raw_gql_result_verify_headers_not_present(client: GitlabClient, requests_mock: requests_mock.Mocker):
    requests_mock.post(url=TEST_BASE_GQL_URL, json={}, request_headers={'Authorization': "A BAD TOKEN"})
    with pytest.raises(NoMockAddress):
        client.get_raw_result_gql(query_body='')
        
def test_page_results_gql_no_next_page(client: GitlabClient, requests_mock: requests_mock.Mocker):
    path_to_page_results = 'data.pages'
    query = """
        {{
            pages(first: %s, cursor: %s) {{
                pageInfo {{ hasNextPage, endCursor }}
                id, name
            }}
        }}
    """
    json_payload = {
            'data': {
                'pages': {
                    'pageInfo': {
                        'hasNextPage': False,
                        'endCursor': '123'
                    },
                    'page': [
                        {
                            'id': 1,
                            'name': 'one'
                        },
                        {
                            'id': 2,
                            'name': 'two'
                        }
                    ]
                }
            }
        }
    requests_mock.post(
        url=TEST_BASE_GQL_URL, 
        request_headers=EXPECTED_AUTH_HEADER,
        json=json_payload,
    )
    results = []
    for result_page in client.page_results_gql(query_body=query, path_to_page_info=path_to_page_results, page_size=2):
        for item in result_page['data']['pages']['page']:
            results.append(item)
            
    assert len(results) == 2
    assert results[0]['id'] == 1
    assert results[1]['id'] == 2

def test_get_organizations_with_gql(client: GitlabClient, requests_mock: requests_mock.Mocker):
    combined_raw_groups = spoof_organizations_through_gql(requests_mock)
    
    organizations = [o for o in client.get_organizations_gql()]
    assert len(organizations) == len(combined_raw_groups)
    for group in organizations:
        assert group in combined_raw_groups

def test_get_repositories_with_gql(client: GitlabClient, requests_mock: requests_mock.Mocker):
    combined_raw_repos = spoof_repositories_through_gql(requests_mock)
    
    repos = [o for o in client.get_repos_gql(group_full_path='')]
    assert len(repos) == len(combined_raw_repos)
    assert len(repos) == 2
    for repo in repos:
        assert repo in combined_raw_repos



import statistics
from unittest.mock import patch

import pytest

from jf_ingest.adaptive_throttler import AdaptiveThrottler


@pytest.fixture
def throttler():
    """Fixture for setting up the AdaptiveThrottler with test parameters."""
    return AdaptiveThrottler(
        max_rps=10.0,
        baseline_window_size=10,
        percentile_threshold=0.90,
        backoff_factor=0.8,
        reverse_backoff_factor=0.95,
    )


def _get_time_values(response_times: list[float]) -> tuple[list[float], float]:
    time_values = []
    cumulative_time = 1000.0

    for rt in response_times:
        time_values.append(cumulative_time)  # Start time
        cumulative_time += rt
        time_values.append(cumulative_time)  # End time

    return time_values, cumulative_time


def _get_threshold_time(response_times: list[float], throttler: AdaptiveThrottler) -> float:
    # Calculate expected percentile threshold time
    sorted_times: list[float] = sorted(response_times)
    percentile_index = int(len(sorted_times) * throttler._percentile_threshold)
    percentile_index = min(percentile_index, len(sorted_times) - 1)
    return sorted_times[percentile_index]


def test_baseline_collection(throttler):
    """Test if the baseline is properly calculated after collecting enough response times."""
    response_times = [0.1, 0.2, 0.15, 0.3, 0.25, 0.2, 0.18, 0.22, 0.2, 0.19]

    # Prepare time values for the baseline
    time_values, cumulative_time = _get_time_values(response_times)
    time_iter = iter(time_values)

    with patch('time.perf_counter', side_effect=lambda: next(time_iter)):
        for _ in response_times:
            with throttler.process_response_time():
                pass

    # Calculate expected 95th percentile time
    sorted_times = sorted(response_times)
    index = int(throttler._percentile_threshold * len(sorted_times))
    percentile_value = sorted_times[index]

    # Calculate standard deviation
    std_dev = statistics.stdev(response_times)

    # Calculate expected adjusted threshold (percentile + 2 * std_dev)
    expected_percentile_time = percentile_value + (2 * std_dev)

    # Calculate expected initial request rate
    average_time = sum(response_times) / len(response_times)
    expected_request_rate = min(1 / average_time, throttler.max_rps)

    # Assert that the adjusted percentile threshold is correct
    assert throttler._percentile_threshold_time == pytest.approx(expected_percentile_time), (
        f"Expected adjusted threshold time {expected_percentile_time}, "
        f"but got {throttler._percentile_threshold_time}"
    )

    # Assert that the initial request rate is correctly calculated
    assert throttler._initial_request_rate == pytest.approx(expected_request_rate), (
        f"Expected initial request rate {expected_request_rate}, "
        f"but got {throttler._initial_request_rate}"
    )


def test_throttling_when_above_threshold(throttler):
    """Test if throttling occurs when response times exceed the threshold."""
    # Initial response times to establish the baseline
    response_times = [0.1, 0.2, 0.15, 0.3, 0.25, 0.2, 0.18, 0.22, 0.2, 0.19]

    # Prepare time values for the baseline
    time_values, cumulative_time = _get_time_values(response_times)

    # Calculate expected percentile threshold time
    expected_percentile_time = _get_threshold_time(response_times, throttler)

    # Simulate a response time above the threshold
    response_above_threshold = expected_percentile_time + 0.15
    time_values.append(cumulative_time)  # Start time of the above-threshold response
    cumulative_time += response_above_threshold
    time_values.append(cumulative_time)  # End time

    time_iter = iter(time_values)

    with (
        patch('time.perf_counter', side_effect=lambda: next(time_iter)),
        patch('time.sleep') as mock_sleep,
    ):
        # Establish baseline
        for _ in response_times:
            with throttler.process_response_time():
                pass

        assert throttler._percentile_threshold_time > 0, "Percentile threshold time should be set."
        assert throttler._request_rate > 0, "Initial request rate should be calculated."

        # Capture initial rate
        initial_rate = throttler._request_rate
        expected_throttled_rate = initial_rate * throttler.backoff_factor
        expected_backoff_time = 1 / expected_throttled_rate

        # Simulate response time that exceeds the threshold
        with throttler.process_response_time():
            pass

        # Check that time.sleep was called with the expected backoff time
        mock_sleep.assert_called_with(expected_backoff_time)

        # Assertions with calculated expected values
        assert throttler._request_rate == expected_throttled_rate, (
            f"Expected throttled rate {expected_throttled_rate}, "
            f"but got {throttler._request_rate}"
        )


def test_recovery_when_below_threshold(throttler):
    """Test if recovery occurs incrementally when response times are below the threshold."""
    # Initial response times to establish the baseline
    response_times = [0.1, 0.2, 0.15, 0.2, 0.18, 0.22, 0.19, 0.2, 0.25, 0.24]

    # Prepare time values for the baseline
    time_values, cumulative_time = _get_time_values(response_times)

    # Calculate expected percentile threshold time
    expected_percentile_time = _get_threshold_time(response_times, throttler)

    # Simulate a response above the threshold
    response_above_threshold = expected_percentile_time + 0.1
    time_values.append(cumulative_time)  # Start time of the above-threshold response
    cumulative_time += response_above_threshold
    time_values.append(cumulative_time)  # End time

    # Simulate recovery responses
    total_cycles = 2
    for _ in range(total_cycles * throttler.reverse_backoff_counter):
        response_below_threshold = expected_percentile_time - 0.05
        time_values.append(cumulative_time)
        cumulative_time += response_below_threshold
        time_values.append(cumulative_time)

    time_iter = iter(time_values)

    with patch('time.perf_counter', side_effect=lambda: next(time_iter)), patch('time.sleep'):
        # Establish baseline
        for _ in response_times:
            with throttler.process_response_time():
                pass

        # Trigger throttling
        with throttler.process_response_time():
            pass

        # Simulate recovery with consecutive below-threshold responses
        for cycle in range(1, total_cycles + 1):
            request_rate_before_recovery = throttler._request_rate

            for _ in range(throttler.reverse_backoff_counter):
                with throttler.process_response_time():
                    pass

            expected_recovered_rate = min(
                request_rate_before_recovery / throttler.reverse_backoff_factor,
                throttler._initial_request_rate,
            )

            assert throttler._request_rate == pytest.approx(expected_recovered_rate, rel=1e-5), (
                f"Cycle {cycle}: Expected recovered rate {expected_recovered_rate}, "
                f"but got {throttler._request_rate}"
            )


def test_no_recovery_if_not_enough_consecutive_responses(throttler):
    """
    Test if recovery does not occur if fewer than reverse_backoff_counter
    consecutive responses are within threshold.
    """
    # Initial response times to establish the baseline
    response_times = [0.1, 0.2, 0.15, 0.2, 0.18, 0.22, 0.19, 0.2, 0.25, 0.24]

    # Prepare time values for the baseline
    time_values, cumulative_time = _get_time_values(response_times)

    # Calculate expected percentile threshold time
    expected_percentile_time = _get_threshold_time(response_times, throttler)

    # Simulate a response above the threshold
    response_above_threshold = expected_percentile_time + 0.1
    time_values.append(cumulative_time)  # Start time of the above-threshold response
    cumulative_time += response_above_threshold
    time_values.append(cumulative_time)  # End time

    # Simulate fewer than reverse_backoff_counter consecutive below-threshold responses
    insufficient_responses = throttler.reverse_backoff_counter - 5

    for _ in range(insufficient_responses):
        response_below_threshold = expected_percentile_time - 0.05
        time_values.append(cumulative_time)
        cumulative_time += response_below_threshold
        time_values.append(cumulative_time)

    time_iter = iter(time_values)

    with patch('time.perf_counter', side_effect=lambda: next(time_iter)), patch('time.sleep'):
        # Establish baseline
        for _ in response_times:
            with throttler.process_response_time():
                pass

        # Trigger throttling
        with throttler.process_response_time():
            pass

        initial_rate = throttler._request_rate

        # Simulate responses below threshold but fewer than reverse_backoff_counter
        for _ in range(insufficient_responses):
            with throttler.process_response_time():
                pass

        assert throttler._request_rate == initial_rate, (
            f"Expected rate to remain at {initial_rate} due to insufficient consecutive responses, "
            f"but got {throttler._request_rate}"
        )



from datetime import datetime, timedelta, timezone

import pytest

from jf_ingest.config import GitAuthConfig, GitConfig, GitProvider
from jf_ingest.jf_git.adapters import (
    BackpopulationWindow,
    GitAdapter,
    determine_commit_backpopulation_window,
    determine_pr_backpopulation_window,
)
from jf_ingest.jf_git.adapters.github import GithubAdapter
from jf_ingest.jf_git.standardized_models import (
    StandardizedBranch,
    StandardizedCommit,
    StandardizedOrganization,
    StandardizedRepository,
)

GIT_CONFIG_DEFAULTS = {
    'company_slug': 'test_company',
    'instance_slug': 'test_instance',
    'instance_file_key': 'test_FILEKEY',
    'git_provider': 'GITHUB',
    'git_auth_config': None,
}

GIT_TEST_ORG = StandardizedOrganization(id='test_org', name='Test Org', login='Test-Org', url='')
GIT_TEST_REPO = StandardizedRepository(
    id='1',
    name='test_repo',
    full_name='test_repo_full_name',
    url='',
    is_fork=False,
    default_branch_name='main',
    default_branch_sha='test',
    organization=GIT_TEST_ORG,
)

TEST_GIT_AUTH_CONFIG = GitAuthConfig(
        company_slug='TEST COMPANY',
        base_url='www.base.com',
        token='apples',
        verify=False,
    )
BASE_ORGANIZATION = StandardizedOrganization(
    id='org_login',
    name='An Organizaiton',
    login='org_login',
    url='www.organization.com'
)

ALT_BASE_ORGANIZATION = StandardizedOrganization(
    id='org_login_alt',
    name='An Alternative Organizaiton',
    login='org_login_alt',
    url='www.organization.com'
)

GITHUB_CONFIG = GitConfig(
    company_slug=TEST_GIT_AUTH_CONFIG.company_slug,
    instance_slug=f'{TEST_GIT_AUTH_CONFIG}_slug',
    instance_file_key='TEST_INSTANCE_FILE_KEY',
    git_provider=GitProvider.GITHUB,
    git_auth_config=TEST_GIT_AUTH_CONFIG,
    git_organizations=[BASE_ORGANIZATION.login]
)

@pytest.fixture
def adapter() -> GitAdapter:
    # We can't stand up the abstract GitAdapter class,
    # so for now just use the GithubAdapter child class
    return GithubAdapter(GITHUB_CONFIG)

def create_git_config(*args, **kwargs):
    return GitConfig(*args, **kwargs)

STANDARDIZED_REPOS = [
    StandardizedRepository(
        id='repo_1',
        name='repo_1_name',
        full_name='repository_1_name',
        url='www.repo1.com',
        is_fork=False,
        default_branch_name='main',
        default_branch_sha='',
        organization=BASE_ORGANIZATION
    ),
    StandardizedRepository(
        id='repo_2',
        name='repo_2_name',
        full_name='repository_2_name',
        url='www.repo2.com',
        is_fork=False,
        default_branch_name='develop',
        default_branch_sha='',
        organization=BASE_ORGANIZATION
    ),
    StandardizedRepository(
        id='repo_3',
        name='repo_3_name',
        full_name='repository_3_name',
        url='www.repo3.com',
        is_fork=False,
        default_branch_name='main',
        default_branch_sha='',
        organization=ALT_BASE_ORGANIZATION
    )
]
    

def test_transform_repos_before_saving(adapter: GitAdapter):
    adapter.config.git_redact_names_and_urls = False
    adapter.config.git_strip_text_content = False
    transformed_objects = adapter._transform_data_objects_before_saving(STANDARDIZED_REPOS)
    for transformed_object, object in zip(transformed_objects, STANDARDIZED_REPOS):
        assert object.id == transformed_object['id']
        assert object.name == transformed_object['name']
        assert object.full_name == transformed_object['full_name']
        assert object.url == transformed_object['url']
        assert object.is_fork == transformed_object['is_fork']
        assert object.default_branch_sha == transformed_object['default_branch_sha']
        assert object.default_branch_name == transformed_object['default_branch_name']
        assert object.organization.name == transformed_object['organization']['name']
        assert object.organization.url == transformed_object['organization']['url']
    

def test_transform_repos_before_saving_with_strip_text(adapter: GitAdapter):
    adapter.config.git_redact_names_and_urls = False
    adapter.config.git_strip_text_content = True
    transformed_objects = adapter._transform_data_objects_before_saving(STANDARDIZED_REPOS)
    for transformed_object, object in zip(transformed_objects, STANDARDIZED_REPOS):
        assert object.id == transformed_object['id']
        assert object.name == transformed_object['name']
        assert object.full_name == transformed_object['full_name']
        assert object.url == transformed_object['url']
        assert object.is_fork == transformed_object['is_fork']
        assert object.default_branch_sha == transformed_object['default_branch_sha']
        assert object.default_branch_name == transformed_object['default_branch_name']
        assert object.organization.name == transformed_object['organization']['name']
        assert object.organization.url == transformed_object['organization']['url']

def test_transform_repos_before_saving_with_redaction(adapter: GitAdapter):
    adapter.config.git_redact_names_and_urls = True
    adapter.config.git_strip_text_content = False
    transformed_objects = adapter._transform_data_objects_before_saving(STANDARDIZED_REPOS)
    repo_redaction_index = 0
    for (transformed_object, object) in zip(transformed_objects, STANDARDIZED_REPOS):
        assert object.id == transformed_object['id']
        assert f'redacted-000{repo_redaction_index}' == transformed_object['name']
        assert f'redacted-000{repo_redaction_index+1}' == transformed_object['full_name']
        assert None == transformed_object['url']
        assert object.is_fork == transformed_object['is_fork']
        assert object.default_branch_sha == transformed_object['default_branch_sha']
        if object.id == 'repo_3':  # Repo 3 is in a differnt org, so the redaction counter should go up
            assert f'redacted-0001' == transformed_object['organization']['name']
        else:
            assert f'redacted-0000' == transformed_object['organization']['name']
        assert None == transformed_object['organization']['url']
        
        repo_redaction_index += 2

# Test strip text content with commits
STANDARDIZED_COMMITS = [
    StandardizedCommit(
        hash='1234',
        url='www.commit-1.com',
        message='OJ-12342: new commit',
        commit_date=datetime(2024, 1, 1),
        author=datetime(2024, 1, 1),
        author_date=None,
        repo=STANDARDIZED_REPOS[0],
        is_merge=False,
        branch_name='main'
    ),
    StandardizedCommit(
        hash='12345',
        url='www.commit-2.com',
        message='OJ-60845515: new commit',
        commit_date=datetime(2024, 1, 1),
        author=datetime(2024, 1, 1),
        author_date=None,
        repo=STANDARDIZED_REPOS[-1],
        is_merge=False,
        branch_name='master'
    ),
]

def test_transform_commits_before_saving(adapter: GitAdapter):
    adapter.config.git_redact_names_and_urls = False
    adapter.config.git_strip_text_content = False
    transformed_objects = adapter._transform_data_objects_before_saving(STANDARDIZED_COMMITS)
    for transformed_object, standardized_object in zip(transformed_objects, STANDARDIZED_COMMITS):
        assert transformed_object['hash'] == standardized_object.hash
        assert transformed_object['url'] == standardized_object.url
        assert transformed_object['message'] == standardized_object.message
        assert transformed_object['commit_date'] == standardized_object.commit_date
        assert transformed_object['author_date'] == standardized_object.author_date
        assert transformed_object['author'] == standardized_object.author
        assert transformed_object['is_merge'] == standardized_object.is_merge
        assert transformed_object['branch_name'] == standardized_object.branch_name
        # short repo fields
        assert transformed_object['repo']['id'] == standardized_object.repo.id
        assert transformed_object['repo']['name'] == standardized_object.repo.name
        assert transformed_object['repo']['full_name'] == standardized_object.repo.full_name
        assert transformed_object['repo']['url'] == standardized_object.repo.url
        # Org fields
        assert transformed_object['repo']['organization']['login'] == standardized_object.repo.organization.login
        assert transformed_object['repo']['organization']['url'] == standardized_object.repo.organization.url
        assert transformed_object['repo']['organization']['name'] == standardized_object.repo.organization.name
        
        
def test_transform_commits_before_saving_with_redact_and_strip(adapter: GitAdapter):
    adapter.config.git_redact_names_and_urls = True
    adapter.config.git_strip_text_content = True
    transformed_objects = adapter._transform_data_objects_before_saving(STANDARDIZED_COMMITS)
    branch_redaction_index = 0
    repo_redaction_index = 0
    organization_redaction_index = 0
    for transformed_object, standardized_object in zip(transformed_objects, STANDARDIZED_COMMITS):
        assert transformed_object['hash'] == standardized_object.hash
        assert transformed_object['url'] == None
        assert 'OJ-60845515' == transformed_object['message'] or transformed_object['message'] == 'OJ-12342'
        assert transformed_object['commit_date'] == standardized_object.commit_date
        assert transformed_object['author_date'] == standardized_object.author_date
        assert transformed_object['author'] == standardized_object.author
        assert transformed_object['is_merge'] == standardized_object.is_merge
        assert transformed_object['branch_name'] == f'redacted-000{branch_redaction_index}' or transformed_object['branch_name'] == 'master'
        # short repo fields
        assert transformed_object['repo']['id'] == standardized_object.repo.id
        assert transformed_object['repo']['name'] == f'redacted-000{repo_redaction_index}'
        assert transformed_object['repo']['full_name'] == f'redacted-000{repo_redaction_index + 1}'
        assert transformed_object['repo']['url'] == None
        # Orgs
        assert transformed_object['repo']['organization']['login'] == standardized_object.repo.organization.login and 'redacted' not in standardized_object.repo.organization.login
        assert transformed_object['repo']['organization']['url'] == None
        assert transformed_object['repo']['organization']['name'] == f'redacted-000{organization_redaction_index}'
        repo_redaction_index += 4
        organization_redaction_index += 1
        branch_redaction_index += 1

def test_backpopulation_window_helper_confirm_always_none_cases():
    pull_from = datetime(2024, 1, 1)
    object_last_pulled_date = pull_from

    git_config_args = {
        'pull_from': pull_from,
        'force_full_backpopulation_pull': True,
        'backpopulation_window_days': 10,
        'repos_to_prs_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'repos_to_commits_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window == None
    assert prs_back_population_window == None

    ######################################################
    git_config_args = {
        'pull_from': pull_from,
        'force_full_backpopulation_pull': True,
        'backpopulation_window_days': 10,
        'repos_to_prs_backpopulated_to': {
            GIT_TEST_REPO.id: object_last_pulled_date - timedelta(days=1)
        },
        'repos_to_commits_backpopulated_to': {
            GIT_TEST_REPO.id: object_last_pulled_date - timedelta(days=1)
        },
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window == None
    assert prs_back_population_window == None


def test_backpopulation_window_helper_with_force_pull():
    pull_from = datetime(2024, 1, 1)
    object_last_pulled_date = datetime(2024, 2, 1)
    git_config_args = {
        'pull_from': pull_from,
        'repos_to_prs_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'force_full_backpopulation_pull': True,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert prs_back_population_window == BackpopulationWindow(pull_from, object_last_pulled_date)

    #############################
    git_config_args = {
        'pull_from': pull_from,
        'repos_to_commits_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'force_full_backpopulation_pull': True,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window == BackpopulationWindow(
        pull_from, object_last_pulled_date
    )


def test_backpopulation_window_helper_with_custom_days_window():
    pull_from = datetime(2024, 1, 1, tzinfo=timezone.utc)
    object_last_pulled_date = datetime(2024, 2, 1, tzinfo=timezone.utc)
    backpop_window_days = 10
    target_backpopulation_start_date = object_last_pulled_date - timedelta(days=backpop_window_days)
    git_config_args = {
        'pull_from': pull_from,
        'repos_to_prs_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'backpopulation_window_days': backpop_window_days,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert prs_back_population_window == BackpopulationWindow(
        target_backpopulation_start_date, object_last_pulled_date
    )

    #############################
    git_config_args = {
        'pull_from': pull_from,
        'repos_to_commits_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'backpopulation_window_days': backpop_window_days,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window == BackpopulationWindow(
        target_backpopulation_start_date, object_last_pulled_date
    )

    ##############################
    git_config_args = {
        'pull_from': pull_from,
        'repos_to_prs_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'repos_to_commits_backpopulated_to': {GIT_TEST_REPO.id: object_last_pulled_date},
        'backpopulation_window_days': backpop_window_days,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window == BackpopulationWindow(
        target_backpopulation_start_date, object_last_pulled_date
    )
    assert prs_back_population_window == BackpopulationWindow(
        target_backpopulation_start_date, object_last_pulled_date
    )


def test_backpopulation_window_when_there_is_no_data_in_jellyfish():
    pull_from = datetime(2024, 1, 1, tzinfo=timezone.utc)
    start_time = datetime.now().astimezone(timezone.utc)
    backpop_window_days = 10
    git_config_args = {
        'pull_from': pull_from,
        'backpopulation_window_days': backpop_window_days,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window.backpopulation_window_end >= start_time
    assert prs_back_population_window.backpopulation_window_end >= start_time
    assert (
        prs_back_population_window.backpopulation_window_end
        - prs_back_population_window.backpopulation_window_start
    ).days == 10
    assert (
        commits_back_population_window.backpopulation_window_end
        - commits_back_population_window.backpopulation_window_start
    ).days == 10

    #####################################################
    git_config_args = {
        'pull_from': pull_from,
        'force_full_backpopulation_pull': True,
        **GIT_CONFIG_DEFAULTS,
    }
    config = create_git_config(**git_config_args)
    repo = GIT_TEST_REPO

    prs_back_population_window = determine_pr_backpopulation_window(config=config, repo=repo)
    commits_back_population_window = determine_commit_backpopulation_window(
        config=config, repo=repo
    )
    assert commits_back_population_window.backpopulation_window_end >= start_time
    assert prs_back_population_window.backpopulation_window_end >= start_time
    assert prs_back_population_window.backpopulation_window_start == pull_from
    assert commits_back_population_window.backpopulation_window_start == pull_from



from jf_ingest.config import IngestionConfig, JiraDownloadConfig
from jf_ingest.constants import Constants
from jf_ingest.jf_jira import load_issues_in_batches
from tests.test_file_operations import TestIngestIOHelper
from unittest.mock import patch

SPOOF_INGEST_CONFIG = IngestionConfig(company_slug='Test', jellyfish_api_token='', jira_config=JiraDownloadConfig('', '', False))

def get_ingest_io_helper():
    test_ingest_io_helper = TestIngestIOHelper()
    test_ingest_io_helper.setup_method()
    return test_ingest_io_helper.ingest_io_helper
    

def test_load_issues_in_batches_base_line():
    issues = [
        {'id': '1', 'key': 'issue_1', 'fields': {}},
        {'id': '2', 'key': 'issue_2', 'fields': {}},
        {'id': '3', 'key': 'issue_3', 'fields': {}},
        {'id': '4', 'key': 'issue_4', 'fields': {}},
    ]
    ingest_io_helper = get_ingest_io_helper()
    ingest_config = SPOOF_INGEST_CONFIG
    def _generate_issues():
        for issue in issues:
            yield issue
            
    issues_generator = _generate_issues()
    issue_download_info = load_issues_in_batches(
        issues_to_download=issues_generator,
        ingest_io_helper=ingest_io_helper,
        ingest_config=ingest_config,
        jira_config=ingest_config.jira_config,
        batch_number_start=0
    )
    
    for issue in issues:
        assert issue['id'] in issue_download_info.downloaded_ids
        
    assert issue_download_info.issue_ids_too_large_to_upload == set()
    assert issue_download_info.total_batches == 1
    assert issue_download_info.discovered_parent_ids == set()

def test_load_issues_in_batches_filter_out_issue_size():
    JIRA_ISSUE_SIZE_LIMIT = 10000
    issues = [
        {'id': '1', 'key': 'issue_1', 'fields': {}},
        {'id': '2', 'key': 'issue_2', 'fields': {}},
        {'id': '3', 'key': 'issue_3', 'fields': {}},
        {'id': '4', 'key': 'issue_4', 'fields': {}},
        {'id': '5', 'key': 'JUMBO_ISSUE', 'fields': {}, 'changelogs': ['a' for x in range(0, JIRA_ISSUE_SIZE_LIMIT)]}
    ]
    ingest_io_helper = get_ingest_io_helper()
    ingest_config = SPOOF_INGEST_CONFIG
    def _generate_issues():
        for issue in issues:
            yield issue
    
    with patch('jf_ingest.constants.Constants.JIRA_ISSUE_SIZE_LIMIT', new=JIRA_ISSUE_SIZE_LIMIT):
        issues_generator = _generate_issues()
        issue_download_info = load_issues_in_batches(
            issues_to_download=issues_generator,
            ingest_io_helper=ingest_io_helper,
            ingest_config=ingest_config,
            jira_config=ingest_config.jira_config,
            batch_number_start=0
        )
        
        assert len(issue_download_info.downloaded_ids) == 4
        assert '1' in issue_download_info.downloaded_ids
        assert '2' in issue_download_info.downloaded_ids
        assert '3' in issue_download_info.downloaded_ids
        assert '4' in issue_download_info.downloaded_ids
        assert '5' not in issue_download_info.downloaded_ids
        assert '5' in issue_download_info.issue_ids_too_large_to_upload


def test_load_issues_in_batches_with_parents():
    issues = [
        {'id': '1', 'key': 'issue_1', 'fields': {}},
        {'id': '2', 'key': 'issue_2', 'fields': {}},
        {'id': '3', 'key': 'issue_3', 'fields': {}},
        {'id': '4', 'key': 'issue_4', 'fields': {'parent': {'id': '5'}}},
        {'id': '5', 'key': 'issue_5', 'fields': {'parent': {'id': '6'}}}
    ]
    ingest_io_helper = get_ingest_io_helper()
    ingest_config = SPOOF_INGEST_CONFIG
    def _generate_issues():
        for issue in issues:
            yield issue

    issues_generator = _generate_issues()
    issue_download_info = load_issues_in_batches(
        issues_to_download=issues_generator,
        ingest_io_helper=ingest_io_helper,
        ingest_config=ingest_config,
        jira_config=ingest_config.jira_config,
        batch_number_start=0
    )
    
    assert len(issue_download_info.downloaded_ids) == len(issues)
    assert '1' in issue_download_info.downloaded_ids
    assert '2' in issue_download_info.downloaded_ids
    assert '3' in issue_download_info.downloaded_ids
    assert '4' in issue_download_info.downloaded_ids
    assert '5' in issue_download_info.downloaded_ids
    assert '5' in issue_download_info.discovered_parent_ids
    assert '6' in issue_download_info.discovered_parent_ids
    assert '6' not in issue_download_info.downloaded_ids


class JiraRetryLimitExceeded(Exception):
    pass


class NoAccessibleProjectsException(Exception):
    pass


class NoJiraUsersFoundException(Exception):
    pass


class AtlassianConnectException(Exception):
    pass
